##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
States related to TPC voting.

The implementation of locking the database for commit, flushing
temporary objects to the database, and moving them to their final locations,
live here.
"""
from __future__ import absolute_import
from __future__ import print_function

import time

from zope.interface import implementer

from ZODB.ConflictResolution import ConflictResolvingStorage
from ZODB.POSException import ConflictError
from ZODB.POSException import StorageTransactionError
from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64

from relstorage._util import log_timed
from relstorage._util import log_timed_only_self
from relstorage._util import do_log_duration_info
from relstorage._util import TRACE
from ..interfaces import VoteReadConflictError
from ..interfaces import ITPCStateVoting

from . import LOCK_EARLY
from . import AbstractTPCStateDatabaseAvailable
from .finish import Finish


logger = __import__('logging').getLogger(__name__)
perf_logger = logger.getChild('timing')

class DatabaseLockedForTid(object):

    @classmethod
    def lock_database_for_next_tid(cls, cursor, adapter, ude):
        # We're midway between the state of a database-wide lock
        # and consistent row-level locking. The lock here is now
        # a row-level artificial lock on COMMIT_ROW_LOCK, and we then
        # read TRANSACTION (or OBJECT_STATE in HF).
        # TODO: Continue working to remove the need for the artificial
        # lock.
        user, desc, ext = ude
        tid_int = adapter.lock_database_and_choose_next_tid(
            cursor,
            user,
            desc,
            ext
        )

        tid = int64_to_8bytes(tid_int)
        return cls(tid, tid_int, adapter)

    @classmethod
    def lock_database_for_given_tid(cls, tid, tid_is_packed,
                                    cursor, adapter, ude):
        adapter.locker.hold_commit_lock(cursor, ensure_current=(tid,))
        tid_int = bytes8_to_int64(tid)
        user, desc, ext = ude
        adapter.txncontrol.add_transaction(
            cursor, tid_int, user, desc, ext, tid_is_packed)
        return cls(tid, tid_int, adapter)

    __slots__ = (
        'tid',
        'tid_int',
        'release_commit_lock',
        'local_allocation_time',
    )

    def __init__(self, tid, tid_int, adapter):
        self.tid = tid
        self.tid_int = tid_int
        self.release_commit_lock = adapter.locker.release_commit_lock
        self.local_allocation_time = time.time()

    def __repr__(self):
        return "<%s tid_int=%d created=%s>" %(
            self.__class__.__name__,
            self.tid_int,
            self.local_allocation_time
        )

class AbstractVote(AbstractTPCStateDatabaseAvailable):
    """
    The state we're in following ``tpc_vote``.

    Unlike the begin states, you *must* explicitly call :meth:`enter`
    on this object after it is constructed.

    """

    __slots__ = (
        # (user, description, extension) from the transaction. byte objects.
        'ude',
        # required_tids: {oid_int: tid_int}; confirms that certain objects
        # have not changed at commit. May be a BTree
        'required_tids',
        # The DatabaseLockedForTid object
        'committing_tid_lock',
        # {oid_bytes}: Things that get changed as part of the vote process
        # and thus need to be invalidated. TODO: Move to shared state?
        'invalidated_oids',
        # How many conflicts there were to resolve. None if we're not there yet.
        'count_conflicts',
        # The timestamp we gained control after locking, and then the
        # timestamp we completed voting. If it takes "too long" to get
        # around to finishing, we'll log a warning.
        'lock_and_vote_times'
    )

    def __init__(self, begin_state, committing_tid_lock=None):
        # type: (AbstractBegin, DatabaseLockedForTid) -> None

        # If committing_tid is passed to this method, it means the
        # database has already been locked and the TID is locked in.
        # This is (only!) done when we're restoring transactions.
        super(AbstractVote, self).__init__(begin_state.shared_state)

        self.required_tids = begin_state.required_tids or {} # type: Dict[int, int]
        self.ude = begin_state.ude
        self.committing_tid_lock = committing_tid_lock # type: Optional[DatabaseLockedForTid]
        self.count_conflicts = None
        self.lock_and_vote_times = [None, None]

        # Anything that we've undone or deleted is also invalidated.
        self.invalidated_oids = begin_state.invalidated_oids or set() # type: Set[bytes]

    def _tpc_state_extra_repr_info(self):
        return {
            'share_lock_count': len(self.required_tids),
            'conflict_count': self.count_conflicts,
            'invalidated_count': len(self.invalidated_oids),
        }

    def enter(self, storage):
        resolved_in_vote_oid_ints = self._vote(storage)
        self.invalidated_oids.update({int64_to_8bytes(i) for i in resolved_in_vote_oid_ints})
        self.lock_and_vote_times[1] = time.time()

    @log_timed
    def _flush_temps_to_db(self, cursor):
        if self.shared_state.has_temp_data():
            # Don't bother if we're empty.
            self.shared_state.adapter.mover.store_temps(cursor, self.shared_state.temp_storage)

    def __enter_critical_phase_until_transaction_end(self):
        self.shared_state.load_connection.enter_critical_phase_until_transaction_end()
        self.shared_state.store_connection.enter_critical_phase_until_transaction_end()


    def _vote(self, storage):
        """
        Prepare the transaction for final commit.

        Locks (only!) the rows that will be updated or were marked as
        explicit dependencies through
        `checkCurrentSerialInTransaction`, and then verfies those
        dependencies and resolves conflicts.

        If we're using a shared blob dir, we then take out the commit
        lock, in order to move blobs into final position. (TODO: That
        might not be fully necessary, so long as we can properly roll
        back blobs.) Otherwise, this only locks the rows we will impact.
        """
        # It is assumed that self._lock.acquire was called before this
        # method was called.
        store_connection = self.shared_state.store_connection
        cursor = store_connection.cursor
        __traceback_info__ = store_connection, cursor
        assert cursor is not None
        adapter = self.shared_state.adapter

        # execute all remaining batch store operations.
        # This exists as an extension point.
        self._flush_temps_to_db(cursor)

        # Reserve all OIDs used by this transaction.
        # In a typical transaction, this is our first write to the database
        # that's not temporary.

        # In the common case where we allocated OIDs for new objects,
        # the most recent OID we allocated will match the
        # ``max_stored_oid```. In the also-common case where we've
        # *never* allocated objects and we're just updating
        # pre-existing objects, every OID we see will be newer than
        # what we allocated (0, we never allocated). So it is the
        # responsibility of the ``relstorage.storage.oid.OIDs`` object
        # to keep track of seen OIDs across transacations, keeping
        # track of the maximum that it's seen, and only actually
        # sending a request down to the database when it's a genuinely
        # higher OID.
        #
        # The other time this can come up is ``copyTransactionsFrom()`` or ``restore()``:
        # those OIDs originated elsewhere and have no relation to our database sequence. That's
        # ok, the same logic applies.
        #
        # In this way, we minimize trips to the DB. We could be a
        # *little* bit smarter and track whether one of those APIs was
        # used, or whether we're updating existing objects and avoid a
        # bit more overhead, but benchmarking suggests that it's not
        # worth it in common cases.
        storage._oids.set_min_oid(
            store_connection,
            self.shared_state.temp_storage.max_stored_oid)

        # Lock objects being modified and those registered with
        # readCurrent(). This could raise ReadConflictError or locking
        # errors. See ``IRelStorageAdapter`` for details.
        conflicts = adapter.lock_objects_and_detect_conflicts(cursor, self.required_tids)
        self.lock_and_vote_times[0] = time.time()
        # Ok, we have now taken database locks: exclusive for each old
        # object we are updating and shared for each we wanted to
        # read-current. From now through ``tpc_finish`` we should do
        # our best to avoid anything that introduces unpredictable,
        # arbitrary latencies or delays. This is especially the case
        # if we're gevent monkey-patched: switching greenlets could
        # take an arbitrary amount of time to get back to *this*
        # greenlet so we can finish the job and release our locks.
        # That means we should limit opportunities that involve IO
        # (database queries) or even taking internal locks that may be
        # contended. We ask both our load and store connections to
        # boost their priority until commit so that any queries we do
        # make get serviced ASAP (some gevent drivers can actually
        # guarantee this), but only right before we're about to do something
        # that could potentially be unbounded or allow switching.
        invalidated_oid_ints = self.__check_and_resolve_conflicts(storage, conflicts)

        blobs_must_be_moved_now = False
        committing_tid_bytes = None
        if self.committing_tid_lock:
            # We've already picked a TID. Must have called undo().
            committing_tid_bytes = self.committing_tid_lock.tid

        if self.shared_state.has_blobs():
            # Avoid accessing the actual blobhelper unless we need it
            try:
                self.shared_state.blobhelper.vote(committing_tid_bytes)
            except StorageTransactionError:
                # If this raises an STE, it must be a shared (non-db)
                # blobhelper, and the TID must not be locked.
                assert committing_tid_bytes is None
                blobs_must_be_moved_now = True

        if blobs_must_be_moved_now or LOCK_EARLY:
            logger.log(TRACE, "Locking early (for blobs? %s)", blobs_must_be_moved_now)
            # It is crucial to do this only after locking the current
            # object rows in order to prevent deadlock. (The same order as a regular
            # transaction, just slightly sooner.)
            self._lock_and_move(vote_only=True)

        # New storage protocol
        return invalidated_oid_ints

    @log_timed_only_self
    def __check_and_resolve_conflicts(self, storage, conflicts):
        """
        Either raises an `ConflictError`, or successfully resolves
        all conflicts.

        Returns a set of int OIDs for objects modified in this transaction
        but which were then updated by conflict resolution and so must
        be invalidated.

        All the rows needed for detecting conflicts should be locked against
        concurrent changes.

        :param conflicts: A sequence of information needed for detecting
           and resolving conflicts:
           ``(oid_int, committed_tid_int, tid_this_txn_saw_int, committed_state)``.
           If ``tid_this_txn_saw_int`` is None, it was a read-current check,
           and unless the ``committed_tid_int`` matches the expected value,
           a conflict error is raised.
        """
        # pylint:disable=too-many-locals
        invalidated_oid_ints = set()
        if not conflicts:
            return invalidated_oid_ints

        self.count_conflicts = count_conflicts = len(conflicts)

        # In the past, we didn't load all conflicts from the DB at
        # once, just one at a time. This was because we also fetched
        # the new state data from the DB, and it could be large (if
        # lots of conflicts). But now we use the state we have in our
        # local temp cache for the new state, so we don't need to
        # fetch it, meaning this result will be small...
        #
        # ...almost. The resolution process needs three pickles: the
        # one we tried to save, the one we're based off of, and the
        # one currently committed. Remember we have locked objects at
        # this point, so we need to finish ASAP to not block other
        # transactions; in gevent, we need to also avoid giving up
        # control to the event loop for arbitrary periods of time too
        # as it could take a long time to get back to us.

        # - The one we tried to save (the new one) is passed as a
        # parameter. We read this from our local storage, which is
        # probably in memory and thus fast.
        #
        # - The one currently committed can optionally be passed, and
        # if not, loadSerial() is used to get it. It seems somewhat
        # unlikely that it's not in the local pickle cache, so we
        # probably benefit from returning it in the conflict query.
        #
        # - The one we were based off of is always loaded with
        # loadSerial(). We *possibly* have the one we're based off of
        # already in our storage cache, but there's no guarantee. So
        # it's best to prefetch all these things in order to limit the
        # number of database round-trips and the opportunity to block
        # for arbitrary periods of time.
        logger.debug("Attempting to resolve %d conflicts", count_conflicts)

        required_tids = self.required_tids
        old_states_to_prefetch = []
        actual_conflicts = []
        # First, go through and distinguish read-current conflicts from
        # state conflicts (if the adapter didn't do that already).
        for conflict in conflicts:
            oid_int, committed_tid_int, tid_this_txn_saw_int, _ = conflict
            if tid_this_txn_saw_int is not None:
                # An actual conflict. We need the state.
                actual_conflicts.append(conflict)
                old_states_to_prefetch.append((oid_int, tid_this_txn_saw_int))
            else:
                # A readCurrent entry. Did it conflict?
                # Note that some database adapters (MySQL) may have already raised a
                # UnableToLockRowsToReadCurrentError indicating a conflict. That's a type
                # of ReadConflictError like this.
                expect_tid_int = required_tids[oid_int]
                if committed_tid_int != expect_tid_int:
                    raise VoteReadConflictError(
                        oid=int64_to_8bytes(oid_int),
                        serials=(int64_to_8bytes(committed_tid_int),
                                 int64_to_8bytes(expect_tid_int)))

        if not actual_conflicts:
            # Nothing to prefetch or resolve. No need to go critical,
            # we have no other opportunities to switch.
            return invalidated_oid_ints

        # We're probably going to need to make a database query. Elevate our
        # priority and regain control ASAP.
        self.__enter_critical_phase_until_transaction_end()

        old_states_and_tids = self.shared_state.cache.prefetch_for_conflicts(
            self.shared_state.load_connection.cursor,
            old_states_to_prefetch
        )

        tryToResolveConflict = _CachedConflictResolver(
            storage, old_states_and_tids
        ).tryToResolveConflict

        adapter = self.shared_state.adapter
        read_temp = self.shared_state.temp_storage.read_temp
        store_temp = self.shared_state.temp_storage.store_temp

        # The conflicts can be very large binary strings, no need to include
        # them in traceback info. (Plus they could be sensitive.)
        __traceback_info__ = count_conflicts, invalidated_oid_ints

        for conflict in actual_conflicts:
            # Match the names of the arguments used
            oid_int, committed_tid_int, tid_this_txn_saw_int, committedData = conflict

            oid = int64_to_8bytes(oid_int)
            committedSerial = int64_to_8bytes(committed_tid_int)
            oldSerial = int64_to_8bytes(tid_this_txn_saw_int)
            newpickle = read_temp(oid_int)

            # Because we're using the _CachedConflictResolver, we can only loadSerial()
            # one state: the ``oldSerial`` state. Therefore the committedData *must* be
            # given.

            resolved_state = tryToResolveConflict(oid, committedSerial, oldSerial,
                                                  newpickle, committedData)

            if resolved_state is None:
                # unresolvable; kill the whole transaction
                raise ConflictError(
                    oid=oid,
                    serials=(oldSerial, committedSerial),
                    data=newpickle,
                )

            # resolved
            invalidated_oid_ints.add(oid_int)
            store_temp(oid_int, resolved_state, committed_tid_int)

        # We resolved some conflicts, so we need to send them over to the database.
        adapter.mover.replace_temps(
            self.shared_state.store_connection.cursor,
            self.shared_state.temp_storage.iter_for_oids(invalidated_oid_ints)
        )

        return invalidated_oid_ints

    @log_timed
    def _lock_and_move(self, vote_only=False):
        # Here's where we take the global commit lock, and
        # allocate the next available transaction id, storing it
        # into history-preserving DBs. But if someone passed us
        # a TID (``restore`` or ``undo``), then it must already be in the DB, and the lock must
        # already be held.
        #
        # If we've prepared the transaction, then the TID must be in the
        # db, the lock must be held, and we must have finished all of our
        # storage actions. This is only expected to be the case when we have
        # a shared blob dir.
        #
        # Returns True if we also committed to the database.
        if self.shared_state.prepared_txn:
            # Already done; *should* have been vote_only.
            assert self.committing_tid_lock, (self.shared_state.prepared_txn,
                                              self.committing_tid_lock)
            return False

        kwargs = {
            'commit': True
        }
        if self.committing_tid_lock:
            kwargs['committing_tid_int'] = self.committing_tid_lock.tid_int
        if vote_only:
            # Must be voting.
            kwargs['commit'] = False
            if self.shared_state.has_blobs():
                # Avoid accessing the blobhelper unless we need it
                blob_meth = self.shared_state.blobhelper.vote
                kwargs['after_selecting_tid'] = lambda tid_int: blob_meth(int64_to_8bytes(tid_int))

        if vote_only \
           or self.shared_state.adapter.DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE:
            # If we're going to have to make two trips to the database, one to lock it and get a
            # tid and then one to commit and release locks, either because we're
            # just voting right now, not committing, or because the database doesn't
            # support doing that in a single operation, we need to go critical and
            # regain control ASAP so we can complete the operation.
            self.__enter_critical_phase_until_transaction_end()

        # Note that this may commit the load_connection and make it not
        # viable for a historical view anymore.
        committing_tid_int, prepared_txn = self.shared_state.adapter.lock_database_and_move(
            self.shared_state.store_connection, self.shared_state.load_connection,
            self.shared_state.has_blobs(),
            self.ude,
            **kwargs
        )

        self.shared_state.prepared_txn = prepared_txn
        committing_tid_lock = self.committing_tid_lock
        assert committing_tid_lock is None or committing_tid_int == committing_tid_lock.tid_int, (
            committing_tid_int, committing_tid_lock)

        log_msg = "Database lock and tid already allocated: %s"
        if committing_tid_lock is None:
            self.committing_tid_lock = DatabaseLockedForTid(
                int64_to_8bytes(committing_tid_int),
                committing_tid_int,
                self.shared_state.adapter
            )
            log_msg = "Adapter locked database and allocated tid: %s"

        logger.log(TRACE, log_msg, self.committing_tid_lock)

        return kwargs['commit']

    @log_timed
    def tpc_finish(self, storage, transaction, f=None):
        if transaction is not self.transaction:
            raise StorageTransactionError(
                "tpc_finish called with wrong transaction")
        try:
            finish_entry = time.time()
            # Handle the finishing. We cannot/must not fail now.
            # TODO: Move most of this into the Finish class/module.
            did_commit = self._lock_and_move()
            if did_commit:
                locks_released = time.time()
            assert self.committing_tid_lock is not None, self


            # The IStorage docs say that f() "must be called while the
            # storage transaction lock is held." We don't really have a
            # "storage transaction lock", just the global database lock,
            # that we want to drop as quickly as possible, so it would be
            # nice to drop the commit lock and then call f(). This
            # probably doesn't really matter, though, as ZODB.Connection
            # doesn't use f().
            #
            # If we called `lock_and_move` for the first time in this
            # method, then the adapter will have been asked to go ahead
            # and commit, releasing any locks it can (some adapters do,
            # some don't). So we may or may not have a database lock at
            # this point.
            if self.shared_state.has_blobs():
                # Avoid accessing the actual blobhelper unless we need it
                assert not self.shared_state.blobhelper.NEEDS_DB_LOCK_TO_FINISH
                try:
                    self.shared_state.blobhelper.finish(self.committing_tid_lock.tid)
                except (IOError, OSError):
                    # If something failed to move, that's not really a problem:
                    # if we did any moving now, we're just a cache.
                    logger.exception(
                        "Failed to update blob-cache; ignoring (will re-download)"
                    )


            if f is not None:
                f(self.committing_tid_lock.tid)
            next_phase = Finish(self, self.committing_tid_lock.tid_int, not did_commit)
            if not did_commit:
                locks_released = time.time()

            locked_duration = locks_released - self.lock_and_vote_times[0]
            between_vote_and_finish = finish_entry - self.lock_and_vote_times[1]
            do_log_duration_info(
                "Objects were locked by %s for %.3fs",
                AbstractVote.tpc_finish.__wrapped__, # pylint:disable=no-member
                self, None,
                locked_duration,
                perf_logger
            )
            do_log_duration_info(
                "Time between vote exiting and %s entering was %.3fs",
                AbstractVote.tpc_finish.__wrapped__, # pylint:disable=no-member
                self, None,
                between_vote_and_finish,
                perf_logger
            )

            return next_phase
        finally:
            self.shared_state.release()


@implementer(ITPCStateVoting)
class HistoryFree(AbstractVote):
    __slots__ = ()


@implementer(ITPCStateVoting)
class HistoryPreserving(AbstractVote):
    __slots__ = ()

    def __init__(self, begin_state):
        # Using undo() requires a new TID, so if we had already begun
        # a transaction by locking the database and allocating a TID,
        # we must preserve that.
        super(HistoryPreserving, self).__init__(begin_state,
                                                begin_state.committing_tid_lock)


class HistoryPreservingDeleteOnly(HistoryPreserving):
    __slots__ = ()

    def _vote(self, storage):
        if self.shared_state.temp_storage and self.shared_state.temp_storage.stored_oids:
            raise StorageTransactionError("Cannot store and delete at the same time.")
        # We only get here if we've deleted objects, meaning we hold their row locks.
        # We only delete objects once we hold the commit lock.
        assert self.committing_tid_lock
        # Holding the commit lock put an entry in the transaction table,
        # but we don't want to bump the TID or store that data.
        self.shared_state.adapter.txncontrol.delete_transaction(
            self.shared_state.store_connection.cursor,
            self.committing_tid_lock.tid_int
        )
        self.lock_and_vote_times[0] = time.time()
        return ()

    def _lock_and_move(self, vote_only=False):
        # We don't do the final commit,
        # we just prepare.
        self.shared_state.prepared_txn = self.shared_state.adapter.txncontrol.commit_phase1(
            self.shared_state.store_connection,
            self.committing_tid_lock.tid_int
        )
        return False


class _CachedConflictResolver(ConflictResolvingStorage):

    def __init__(self, storage, old_states_and_tids):
        self._old_states_and_tids = old_states_and_tids
        self._crs_transform_record_data = storage._crs_transform_record_data
        self._crs_untransform_record_data = storage._crs_untransform_record_data

    def loadSerial(self, oid, serial):
        state, tid = self._old_states_and_tids[bytes8_to_int64(oid)]
        assert bytes8_to_int64(serial) == tid
        return state



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Connection management for the storage layer.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib

from zope.interface import implementer

from .._compat import wraps
from .._util import Lazy
from .._util import get_positive_integer_from_environ
from . import interfaces

logger = __import__('logging').getLogger(__name__)

__all__ = [
    'LoadConnection',
    'StoreConnection',
]


class AbstractManagedConnection(object):

    _NEW_CONNECTION_NAME = None
    _RESTART_NAME = None
    _ROLLBACK_NAME = 'rollback_quietly'

    def __init__(self, connmanager):
        self.connection = None
        self.connmanager = connmanager
        self._cursor = None
        self.active = False
        self._new_connection = getattr(connmanager, self._NEW_CONNECTION_NAME)
        self._restart = getattr(connmanager, self._RESTART_NAME)
        self._rollback = getattr(connmanager, self._ROLLBACK_NAME)

    # Hook functions
    on_opened = staticmethod(lambda conn, cursor: None)
    # on_rolledback can be called with None, None if an error occurred
    # during the rollback.
    on_rolledback = on_opened
    on_first_use = on_opened

    def __bool__(self):
        """
        This object is true if it has an open connection and cursor that
        has previously been accessed and the ``on_first_use`` hook has been
        called and so we've established a database snapshot.

        This is useful to keep from polling a load connection, for example,
        when not explicitly asked for.
        """
        return self.connection is not None and 'cursor' in self.__dict__

    __nonzero__ = __bool__

    def get_cursor(self):
        if not self.active or self._cursor is None:
            # XXX: If we've been explicitly dropped, do we always want to
            # automatically re-open? Probably not; bad things could happen:
            # a load connection could skip into the future, and a store connection
            # could lose temp tables.
            conn, cursor = self.open_if_needed()
            self.on_first_use(conn, cursor)
            self.active = True
        return self._cursor

    cursor = Lazy(get_cursor, 'cursor')

    def enter_critical_phase_until_transaction_end(self):
        """
        Given an already opened connection, cause it to attempt to
        raise it's priority and return results faster.

        This mostly has meaning for gevent drivers.
        """
        self.connmanager.driver.enter_critical_phase_until_transaction_end(
            *self.open_if_needed()
        )

    def exit_critical_phase(self):
        if self.connection is not None:
            self.connmanager.driver.exit_critical_phase(self.connection, self._cursor)

    def drop(self):
        self.active = False
        conn, cursor = self.connection, self._cursor
        self.connection, self._cursor = None, None
        self.__dict__.pop('cursor', None)
        self.connmanager.rollback_and_close(conn, cursor)

    def commit(self, force=False):
        if self.connection is not None:
            self.connmanager.commit(self.connection, self._cursor, force=force)
            self.__dict__.pop('cursor', None)

    def rollback_quietly(self):
        """
        Make the connection inactive and quietly roll it back.

        If an error occurs, drop the connection.
        """
        clean_rollback = True
        self.active = False
        if self.connection is None:
            assert self._cursor is None
            assert 'cursor' not in self.__dict__
            return clean_rollback

        conn = self.connection
        cur = self._cursor
        self.__dict__.pop('cursor', None) # force on_first_use to be called.
        clean_rollback = self._rollback(conn, cur)
        if not clean_rollback:
            self.drop()

        self.on_rolledback(self.connection, self._cursor)
        return clean_rollback

    def open_if_needed(self):
        if self.connection is None or self._cursor is None:
            self.drop()
            self._open_connection()
        return self.connection, self._cursor

    def _open_connection(self):
        """
        Open a new connection, assigning it to ``connection`` and ``cursor``
        """
        new_conn, new_cursor = self._new_connection()
        self.connection, self._cursor = new_conn, new_cursor
        self.on_opened(new_conn, new_cursor)

    @staticmethod
    def __noop(*args):
        "does nothing"

    def _restart_connection(self):
        "Restart just the connection when we have no cursor."

    def restart(self):
        """
        Restart the connection if there is any chance that it has any associated state.
        """
        if not self:
            assert not self.active, self.__dict__
            if self.connection is not None: # But we have no cursor We
                # do this so that if the connection has closed itself
                # (or been closed by a unit test) we can detect that
                # and restart automatically. We only actually do
                # anything there for store connections.
                self._restart_connection()
            return

        # If we've never accessed the cursor, we shouldn't have any
        # state to restart.
        if not self.active and 'cursor' not in self.__dict__:
            return

        self.active = False
        self.restart_and_call(self.__noop)


    def restart_and_call(self, f, *args, **kw):
        """
        Restart the connection (roll it back) and call a function
        after doing this.

        This may drop and re-connect the connection if necessary.

        :param callable f:
            The function to call: ``f(conn, cursor, *args, **kwargs)``.
            May be called up to twice if it raises a disconnected exception
            on the first try.

        :return: The return value of ``f``.
        """
        @wraps(f)
        def callback(conn, cursor, fresh, *args, **kwargs):
            assert conn is self.connection and cursor is self._cursor
            if not fresh:
                # We need to restart.
                needs_rollback = 'cursor' in self.__dict__
                # This could raise a disconnected exception, or a
                # ReplicaClosedException.
                self._restart(conn, cursor, needs_rollback)
                self.on_rolledback(conn, cursor)
            if f == self.on_first_use:
                self.cursor = cursor
            return f(conn, cursor, *args, **kwargs)

        return self.call(callback, True, *args, **kw)

    def call(self, f, can_reconnect, *args, **kwargs):
        """
        Call a function with the cursor, connecting it if needed.
        If a connection is already open, use that without rolling it back.

        Note that this does not count as a first usage and won't invoke
        that callback.

        :param callable f: Function to call
            ``f(conn, cursor, fresh_connection_p, *args, **kwargs)``.
            The function may be called up to twice, if the *fresh_connection_p* is false
            on the first call and a disconnected exception is raised.
        :keyword bool can_reconnect: If True, then we will attempt to reconnect
            the connection and try again if a disconnected exception is raised in *f*. If False,
            we let that exception propagate. For example, if a transaction is in progress,
            set this to false.
        """
        fresh_connection = False
        if self.connection is None or self._cursor is None:
            # We're closed or disconnected. Start a new connection entirely.
            self.drop()
            self._open_connection()
            fresh_connection = True
            # If we just connected no point in trying again.
            can_reconnect = False

        try:
            return f(self.connection, self._cursor, fresh_connection, *args, **kwargs)
        except self.connmanager.driver.disconnected_exceptions as e:
            # XXX: This feels like it's factored wrong.
            if not can_reconnect:
                raise
            logger.warning("Disconnected (%s) when running %s; attempting reconnect",
                           e, f, exc_info=True)
            self.drop()
            try:
                self._open_connection()
            except:
                logger.exception("Reconnect %s failed", self)
                raise
            logger.info("Reconnected %s to run %s", self, f)
            return f(self.connection, self._cursor, True, *args, **kwargs)

    @contextlib.contextmanager
    def isolated_connection(self):
        conn, cursor = self._new_connection()
        try:
            yield cursor
        finally:
            self.connmanager.rollback_and_close(conn, cursor)

    @contextlib.contextmanager
    def server_side_cursor(self):
        conn, _ = self.open_if_needed()
        ss_cursor = self.connmanager.driver.cursor(conn, server_side=True)
        try:
            yield ss_cursor
        finally:
            ss_cursor.close()

    def __repr__(self):
        return "<%s at 0x%x active=%s, conn=%r cur=%r>" % (
            self.__class__.__name__,
            id(self),
            self.active,
            self.connection,
            self._cursor
        )


@implementer(interfaces.IManagedLoadConnection)
class LoadConnection(AbstractManagedConnection):

    __slots__ = ()

    _NEW_CONNECTION_NAME = 'open_for_load'
    _RESTART_NAME = 'restart_load'


@implementer(interfaces.IManagedStoreConnection)
class StoreConnection(AbstractManagedConnection):

    __slots__ = ()

    _NEW_CONNECTION_NAME = 'open_for_store'
    _RESTART_NAME = 'restart_store'
    _ROLLBACK_NAME = 'rollback_store_quietly'

    def begin(self):
        self.connmanager.begin(*self.open_if_needed())

    def _restart_connection(self):
        self.rollback_quietly()

class PrePackConnection(StoreConnection):
    __slots__ = ()
    _NEW_CONNECTION_NAME = 'open_for_pre_pack'


@implementer(interfaces.IManagedDBConnection)
class ClosedConnection(object):
    """
    Represents a permanently closed connection.
    """
    __slots__ = ()

    cursor = None
    connection = None

    def __init__(self, *args):
        "We have no state."

    def drop(self):
        "Does nothing."

    rollback_quietly = drop

    __bool__ = __nonzero__ = lambda self: False

    def isolated_connection(self, *args, **kwargs):
        raise NotImplementedError

    restart_and_call = isolated_connection
    enter_critical_phase_until_transaction_end = isolated_connection


class StoreConnectionPool(object):
    """
    A thread-safe pool of `StoreConnection` objects.

    Connections are opened on demand; opening a connection on demand
    does not block.

    By default, it will keep around a `StoreConnection` for every instance
    of a RelStorage that has ever used one, which ordinarily corresponds to the
    ZODB Connection pool size. It can be made to keep a smaller (but not larger)
    number around by setting ``MAX_STORE_CONNECTIONS_IN_POOL``.
    """

    MAX_STORE_CONNECTIONS_IN_POOL = get_positive_integer_from_environ(
        'RS_MAX_POOLED_STORE_CONNECTIONS',
        None
    )

    def __init__(self, connmanager):
        import threading
        self._lock = threading.Lock()
        self._connmanager = connmanager
        self._connections = []
        self._count = 1
        self._factory = StoreConnection

    # MVCC protocol
    def new_instance(self):
        with self._lock:
            self._count += 1
        return self

    def release(self):
        with self._lock:
            self._count -= 1
            self._shrink()

    def close(self):
        with self._lock:
            self._count = 0
            self._factory = ClosedConnection
            self._shrink()
            self._connections = ()

    @contextlib.contextmanager
    def borrowing(self, commit=False):
        rollback = True
        try:
            conn = self.borrow()
            yield conn
            if commit:
                conn.commit()
                rollback = False
        finally:
            self._replace(conn, rollback)

    def borrow(self):
        conn = None
        with self._lock:
            if self._connections:
                conn = self._connections.pop()

        if conn is None:
            conn = self._factory(self._connmanager)
        else:
            conn.restart()

        conn.begin()
        return conn

    def replace(self, connection):
        self._replace(connection, True)

    def _replace(self, connection, needs_rollback):
        if needs_rollback:
            clean_rollback = connection.rollback_quietly()
        else:
            clean_rollback = True
        if not clean_rollback:
            connection.drop()
        else:
            connection.exit_critical_phase()
            with self._lock:
                self._connections.append(connection)
                self._shrink()

    def _shrink(self):
        # Call while holding the lock, after putting a connection
        # back in the pool.
        # Limits the number of pooled connections to be no more than
        # ``instance_count`` (i.e., one per open RelStorage), or ``MAX_STORE_CONNECTIONS_IN_POOL``,
        # if set and if less than instance_count
        keep_connections = min(self._count, self.MAX_STORE_CONNECTIONS_IN_POOL or self._count)

        while len(self._connections) > keep_connections and self._connections:
            conn = self._connections.pop()
            conn.drop() # TODO: This could potentially be slow? Might
                        # want to do this outside the lock.
            conn.connmanager = None # It can't be opened again.

    def drop_all(self):
        with self._lock:
            while self._connections:
                conn = self._connections.pop()
                conn.drop()

    def hard_close_all_connections(self):
        # Testing only.
        for conn in self._connections:
            conn.connection.close()

    @property
    def pooled_connection_count(self):
        return len(self._connections)

    @property
    def instance_count(self):
        return self._count

class ClosedConnectionPool(object):

    def borrow(self):
        return ClosedConnection()

    def replace(self, connection):
        "Does Nothing"

    def new_instance(self):
        "Does nothing"

    release = new_instance
    close = release
    drop_all = release

    pooled_connection_count = instance_count = 0


class SingleConnectionPool(object):
    __slots__ = ('connection',)

    def __init__(self, connection):
        self.connection = connection

    @contextlib.contextmanager
    def borrowing(self, commit=False): # pylint:disable=unused-argument
        """
        The *commit* parameter is ignored
        """
        yield self.connection



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import

import os
import time

from zope.interface import implementer

from .._compat import metricmethod
from .interfaces import IReplicaSelector


@implementer(IReplicaSelector)
class ReplicaSelector(object):

    # The time at which we checked the config
    _config_checked = 0

    def __init__(self, fn, replica_timeout):
        self.replica_conf = fn
        self.replica_timeout = replica_timeout
        self._read_config()
        self._select(0)
        self._iterating = False
        self._skip_index = None

    def _read_config(self):
        self._config_modified = os.path.getmtime(self.replica_conf)
        self._config_checked = time.time()
        f = open(self.replica_conf, 'r')
        try:
            lines = f.readlines()
        finally:
            f.close()
        replicas = []
        for line in lines:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            replicas.append(line)
        if not replicas:
            raise IndexError(
                "No replicas specified in %s" % self.replica_conf)
        self._replicas = replicas

    def _is_config_modified(self):
        now = time.time()
        if now < self._config_checked + 1:
            # don't check the last mod time more often than once per second
            return False
        self._config_checked = now
        t = os.path.getmtime(self.replica_conf)
        return t != self._config_modified

    def _select(self, index):
        self._current_replica = self._replicas[index]
        self._current_index = index
        if index > 0 and self.replica_timeout:
            self._expiration = time.time() + self.replica_timeout
        else:
            self._expiration = None

    def current(self):
        """Get the current replica."""
        self._iterating = False
        if self._is_config_modified():
            self._read_config()
            self._select(0)
        elif self._expiration is not None and time.time() >= self._expiration:
            self._select(0)
        return self._current_replica

    @metricmethod
    def next(self):
        """Return the next replica to try.

        Return None if there are no more replicas defined.
        """
        if self._is_config_modified():
            # Start over even if iteration was already in progress.
            self._read_config()
            self._select(0)
            self._skip_index = None
            self._iterating = True
        elif not self._iterating:
            # Start iterating.
            self._skip_index = self._current_index
            i = 0
            if i == self._skip_index:
                i = 1
                if i >= len(self._replicas):
                    # There are no more replicas to try.
                    self._select(0)
                    return None
            self._select(i)
            self._iterating = True
        else:
            # Continue iterating.
            i = self._current_index + 1
            if i == self._skip_index:
                i += 1
            if i >= len(self._replicas):
                # There are no more replicas to try.
                self._select(0)
                return None
            self._select(i)

        return self._current_replica




"""OID tree traversal for the garbage collection phase of packing.

Optimized for memory efficiency.  Uses sets of native integers
rather than Python integers because native integers take up a lot less
room in RAM.
"""

from __future__ import absolute_import

import collections
import gc
import logging

import BTrees

from relstorage._compat import iteritems

IIunion32 = BTrees.family32.II.union # pylint:disable=no-member
IISet32 = BTrees.family32.II.Set
IISet64 = BTrees.family64.II.Set
log = logging.getLogger(__name__)


class IISet32X(object):
    """An IISet32 extended with a Python set layer for efficient inserts."""

    def __init__(self):
        self._base = IISet32()
        self._layer = set()

    def add(self, key):
        layer = self._layer
        if key not in layer and key not in self._base:
            layer.add(key)
            if len(layer) > 10000:
                self._apply()

    def _apply(self):
        """Add the layer to the base and reset the layer."""
        if self._layer:
            self._base = IIunion32(self._base, IISet32(self._layer))
            self._layer.clear()

    def __iter__(self):
        self._apply()
        return iter(self._base)

    def __contains__(self, key):
        return key in self._layer or key in self._base


class TreeMarker(object):
    """Finds all OIDs reachable from a set of root OIDs."""

    # This class groups OIDs by their upper 33 bits.  Why 33 instead
    # of 32?  Because IISet and IIBucket are signed, they can not accept
    # positive integers >= (1 << 31). The solution is to simply
    # add an extra grouping bit.
    hi = 0xffffffff80000000  # 33 high bits
    lo = 0x000000007fffffff  # 31 low bits

    def __init__(self):
        # self._refs:
        # {from_oid_hi: {to_oid_hi: IISet64([from_oid_lo << 32 | to_oid_lo])}}
        self._refs = collections.defaultdict(
            lambda: collections.defaultdict(IISet64))
        # self._reachable: {oid_hi: IISet32X}
        self._reachable = collections.defaultdict(IISet32X)
        self.reachable_count = 0

    def add_refs(self, pairs):
        """Add a list of (from_oid, to_oid) reference pairs.

        `from_oid` and `to_oid` must be 64 bit integers.
        """
        refs = self._refs
        hi = self.hi
        lo = self.lo
        for from_oid, to_oid in pairs:
            s = refs[from_oid & hi][to_oid & hi]
            s.add(((from_oid & lo) << 32) | (to_oid & lo))

    def mark(self, oids):
        """Mark specific OIDs and descendants of those OIDs as reachable."""
        hi = self.hi
        lo = self.lo
        pass_count = 1

        # this_pass: {oid_hi: IISet32X}
        this_pass = collections.defaultdict(IISet32X)
        for oid in sorted(oids):
            this_pass[oid & hi].add(int(oid & lo))

        while this_pass:
            gc.collect()
            found, next_pass = self._mark_pass(this_pass)
            log.debug(
                "Found %d more referenced object(s) in pass %d",
                found, pass_count)
            if not found:
                break
            self.reachable_count += found
            pass_count += 1
            this_pass = next_pass

        return pass_count

    def _mark_pass(self, this_pass):
        """Mark OIDs as reachable. Produce an OID set for the next pass.

        Return (found, next_pass), where `found` is the number of
        new OIDs marked and `next_pass` is the collection of OIDs to
        follow in the next pass.
        """
        # pylint:disable=too-many-locals
        # next_pass: {oid_hi: IISet32X}
        next_pass = collections.defaultdict(IISet32X)
        found = 0
        refs = self._refs
        reachable = self._reachable
        lo = self.lo

        for oid_hi, oids_lo in iteritems(this_pass):
            from_reachable_set = reachable[oid_hi]

            for oid_lo in oids_lo:
                if oid_lo in from_reachable_set:
                    # This OID is already known to be reachable.
                    continue

                found += 1
                from_reachable_set.add(oid_lo)

                if oid_hi not in refs:
                    # This OID doesn't reference anything.
                    continue

                # Add the children of this OID to next_pass.
                for to_oid_hi, s in iteritems(refs[oid_hi]):
                    min_key = oid_lo << 32
                    max_key = min_key | 0xffffffff
                    keys = s.keys(min=min_key, max=max_key)
                    if not keys:
                        # No references found here.
                        continue
                    to_reachable_set = reachable[to_oid_hi]
                    next_pass_add = next_pass[to_oid_hi].add
                    for key in keys:
                        child_oid_lo = int(key & lo)
                        if child_oid_lo not in to_reachable_set:
                            next_pass_add(child_oid_lo)

        return found, next_pass

    def free_refs(self):
        """Free the collection of refs to save RAM."""
        self._refs = None
        gc.collect()

    @property
    def reachable(self):
        """Iterate over all the reachable OIDs."""
        for oid_hi, oids_lo in iteritems(self._reachable):
            for oid_lo in oids_lo:
                # Decode the OID.
                yield oid_hi | oid_lo



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A script to assist in the recovery of a history-free storage damaged
by incorrect garbage collection, such as that described in
https://github.com/zodb/relstorage/issues/344.

.. note:: This does not work with multi-databases.

The script takes as input a ZConfig configuration file just like
zodbconvert does:

    - The *source* database is the backup. Objects to be restored will
      come from here.

    - The *destination* database is the damaged database. Objects will
      be restored into this database.

It also needs a list of missing OIDs (oids that generate POSKeyError).
These may or may not exist in the *source* database. This should be
given on stdin, one missing OID per line, in integer ('u64' or
'bytes8_to_int64') format. As a special case, piping the output from
zc.zodbdgc's ``multi-zodb-check-refs`` script is supported. It may
also be given from a file specified with ``--oid-list``.

The strategy uses the same underlying infrastructure that zodbconvert
does, so it is tested and stable, if not necessarily fast. The
difference is that it inserts a filter into the process. It relies on
the fact that a history-free storage can add objects to a transaction,
and commit for the same TID, multiple times.

    1. First, open both databases.

    2. Read all input OIDs into a set.

    3. Ask the source destination to return the current state and tid
       for those OIDs. (TODO: Batches if this could be large.)

    3b. Ask the destination for the same; if anything is present here, it
        should either be the same or a higher TID. Remove duplicates.

    4. Produce an iterator of transaction records that returns this
       data, sorted and grouped by TID. Wrap this in a phony
       storage-like object.

    5. Pass this to destination.copyTransactionsFrom().

    6. Profit.

The original input list forms the starting "roots". Because the missing
objects may themselves refer to other objects which were also
incorrectly garbage collected (because their only referer was this
object) to an arbitrary depth, the above steps from 3 to 5 must be repeated
using as input the list of missing references defined by all the stats collected
in stage 3 of the previous iteration. Care is taken to avoid running into
problems with references that the source database simply doesn't have.
"""
# TODO: Should we disable the taking of the commit lock; we probably
# don't need it.

# TODO: We actually probably don't really need to even preserve the
# transaction ids and group them in batches that way. We could just
# choose the earliest TID and use that for all of them. That could
# help speed wise. But for best fidelity, it's good to restore them.
# That should keep batches small.

# TODO: When we implement https://github.com/zodb/relstorage/issues/275
# and our own check-refs, optionally use that directly.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import logging
import sys
import zlib

from ZODB.BaseStorage import DataRecord
from ZODB.BaseStorage import TransactionRecord
from ZODB.POSException import POSKeyError
from ZODB.blob import is_blob_record
# This function only returns strong same-database references,
# ignoring weak and cross-database references. We shouldn't have
# the latter, and the former were weak so if they're gone that's
# ok.
from ZODB.serialize import referencesf as get_oids_referenced_by_pickle

from relstorage.zodbconvert import open_storages
from relstorage.interfaces import IRelStorage
from relstorage._compat import OID_SET_TYPE as OidSet
from relstorage._compat import OID_OBJECT_MAP_TYPE as OidMap
from relstorage._compat import OidSet_difference
from relstorage._compat import OidSet_discard
from relstorage._util import int64_to_8bytes
from relstorage._util import bytes8_to_int64

logger = logging.getLogger(__name__)

def check_db_compat(storage):
    assert IRelStorage.providedBy(storage)
    assert not storage._options.keep_history

def read_missing_oids(oid_lines):
    """
    Parse lines into oids.

    >>> list(read_missing_oids([
    ...   "!!! Users 0 ?", "POSKeyError: foo",
    ...   "!!! Users 0 ?",
    ...   "!!! Users 1 ?",
    ...   "bad xref name, 1", "bad db",  ]))
    [0, 1]
    """
    result = OidSet()
    for line in oid_lines:
        if line.startswith('bad') or ':' in line:
            continue
        if line.startswith('!!!'):
            # zc.zodbdgc output. bad OID is always the
            # third field.
            try:
                oid = int(line.split()[2])
            except (ValueError, IndexError):
                logger.info("Malformed zc.zodbdgc input: %s", line)
                continue
            result.add(oid)
        else:
            # Just an int
            try:
                oid = int(line)
            except ValueError:
                logger.info("Malformed input: %s", line)
            else:
                result.add(oid)
    if oid_lines != sys.stdin:
        oid_lines.close()
    return result

class MissingObjectStorage(object):
    """
    The thing we pass to copyTransactionsFrom().
    """

    def __init__(self, backup_state_by_oid, tids_to_oids, backup_storage):
        self._states = backup_state_by_oid
        self._transaction_groups = tids_to_oids
        self._backup_storage = backup_storage

    def __getattr__(self, name):
        return getattr(self._backup_storage, name)

    def close(self):
        pass

    def iterator(self):
        return TransactionRecordIterator(self._states, self._transaction_groups)

class TransactionRecordIterator(object):

    def __init__(self, backup_state_by_oid, tids_to_oids):
        self._states = backup_state_by_oid
        self._len = len(tids_to_oids)
        self._transaction_group_iter = iter(tids_to_oids.items())

    def close(self):
        pass

    def __len__(self):
        return self._len

    def __iter__(self):
        return self

    def __next__(self):
        tid, oids = next(self._transaction_group_iter)
        return RelStorageTransactionRecord(tid, oids, self._states)

    next = __next__ # Python 2

class RelStorageTransactionRecord(TransactionRecord):
    def __init__(self, tid_int, oid_ints, states):
        tid = int64_to_8bytes(tid_int)
        self._oid_ints = oid_ints
        self._states = states
        TransactionRecord.__init__(self, tid, 'p', b'', b'', {})

    def __iter__(self):
        logger.info("Restoring %d oids for tid %d",
                    len(self._oid_ints),
                    bytes8_to_int64(self.tid))
        return RecordIterator(self.tid, self._oid_ints, self._states)

class RecordIterator(object):

    def __init__(self, tid, oid_ints, states):
        self._tid = tid
        self._oid_int_iter = iter(oid_ints)
        self._states = states

    def __next__(self):
        oid_int = next(self._oid_int_iter)
        state, _ = self._states[oid_int]
        logger.info("Restoring OID %d at TID %s (state size %d)",
                    oid_int, bytes8_to_int64(self._tid), len(state))
        return DataRecord(
            int64_to_8bytes(oid_int),
            self._tid,
            state,
            None
        )

    next = __next__ # Python 2

def _find_missing_references_from_pickles(destination, pickles, permanently_gone):
    # Return a set of objects missing from the database given the
    # pickle states of other objects.
    # *permanently_gone* is a set of oid byte strings that are
    # known to be missing and shouldn't be investigated and returned.
    oids = []
    for pickle in pickles:
        # Support zc.zlibstorage wrappers.
        if pickle.startswith(b'.z'):
            pickle = zlib.decompress(pickle[2:])
        get_oids_referenced_by_pickle(pickle, oids)

    logger.info(
        "Given %d pickles, there are %d unique references.",
        len(pickles), len(oids)
    )

    missing_oids = OidSet()
    destination.prefetch(oids)
    for oid in oids:
        if oid in permanently_gone:
            continue
        try:
            state, tid = destination.load(oid, b'')
            if is_blob_record(state):
                destination.loadBlob(oid, tid)
        except POSKeyError:
            missing_oids.add(bytes8_to_int64(oid))
    logger.info(
        "Given %d pickles, there are %d missing references.",
        len(pickles), len(missing_oids)
    )
    return missing_oids

def _restore_missing_oids(source, destination, missing_oids, permanently_gone):
    # Implements steps 3 through 5.

    # Returns a set of oid ints that are missing from the database
    # considering only the objects that were restored.

    # 3. Get backup data
    # (oid, state, tid)
    backup_data = source._adapter.mover.load_currents(
        source._load_connection.cursor,
        missing_oids
    )

    # {oid: (state, tid)}
    backup_state_by_oid = OidMap()
    tids_to_oids = OidMap()
    for oid, state, tid in backup_data:
        if tid not in tids_to_oids:
            tids_to_oids[tid] = OidSet()
        tids_to_oids[tid].add(oid)
        backup_state_by_oid[oid] = (state, tid)

    found_oids = OidSet(backup_state_by_oid)
    oids_required_but_not_in_backup = OidSet_difference(missing_oids, found_oids)
    if oids_required_but_not_in_backup:
        logger.warning(
            "The backup storage is missing %d OIDs needed",
            len(oids_required_but_not_in_backup)
        )
        permanently_gone.update(oids_required_but_not_in_backup)
    del found_oids
    del oids_required_but_not_in_backup

    # 3b. Compare with destination.
    current_data = destination._adapter.mover.load_currents(
        destination._load_connection.cursor,
        backup_state_by_oid
    )
    current_data = list(current_data)
    for oid, _, tid in current_data:
        if oid not in backup_state_by_oid:
            # Common, expected case.
            continue

        _, backup_tid = backup_state_by_oid[oid]

        logger.warning(
            "Destination already contains data for %d. Check your OID list. "
            "Refusing to overwrite. (Source TID: %d; Destination TID: %d)",
            oid, backup_tid, tid
        )
        # If we're doing a dry-run, it's probably in here.
        OidSet_discard(permanently_gone, oid)
        del backup_state_by_oid[oid]
        tids_to_oids[backup_tid].remove(oid)
        if not tids_to_oids[backup_tid]:
            del tids_to_oids[backup_tid]
        continue

    if not tids_to_oids:
        logger.warning("All OIDs already present in destination; nothing to do.")
        return

    # 4. Produce phony storage that iterates the backup data.
    logger.info(
        "Beginning restore of %d OIDs in %d transactions.",
        len(backup_state_by_oid),
        len(tids_to_oids)
    )
    copy_from = MissingObjectStorage(backup_state_by_oid, tids_to_oids, source)

    # 5. Hand it over to be copied.
    destination.copyTransactionsFrom(copy_from)

    # Find anything still missing, after having stored what we could.
    newly_missing = _find_missing_references_from_pickles(
        destination,
        [x[0] for x in backup_state_by_oid.values()],
        permanently_gone)
    return newly_missing

def main(argv=None):
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(name)s] %(levelname)s %(message)s")

    argv = argv if argv is not None else sys.argv

    # 1. Open the source and destination.
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--dry-run", dest="dry_run", action="store_true",
        default=False,
        help="Attempt to open both storages, then explain what would be done."
    )
    parser.add_argument(
        '--oid-list', dest='oid_list', type=argparse.FileType('r'),
        default='-',
        help="Where to read the list of OIDs from. "
             "Defaults to stdin for use in a pipe from zc.zodbgc."
    )
    parser.add_argument("config_file", type=argparse.FileType('r'))

    options = parser.parse_args(argv[1:])
    permanently_gone = OidSet()

    source, destination = open_storages(options)
    check_db_compat(source)
    check_db_compat(destination)

    if options.dry_run:
        # Make sure we can't commit.
        destination.tpc_finish = destination.tpc_abort


    # 2. Read incoming OIDs.
    missing_oids = read_missing_oids(options.oid_list)
    if not missing_oids:
        sys.exit("Unable to read any missing OIDs.")

    if options.dry_run:
        # And since we can't commit, the starting set of objects
        # known to be permanently gone are...the set we start with!
        permanently_gone.update(missing_oids)

    try:
        while missing_oids:
            missing_oids = _restore_missing_oids(source, destination,
                                                 missing_oids, permanently_gone)
    finally:
        source.close()
        destination.close()

    if permanently_gone:
        logger.warning(
            "The following referenced OIDs could not be recovered: %s",
            list(permanently_gone)
        )

if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Helpers for various disk-based persistent storage format.

Doesn't actually do any persistence itself.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import errno
import logging
import os
import os.path
import sqlite3


from relstorage.adapters.sqlite.drivers import Sqlite3Driver
from relstorage.adapters.sqlite.dialect import SQ3_SUPPORTS_CTE

# Because we use a CTE in our default queries. In principle,
# we could probably re-write queries and run on 3.7.11; 3.7.0 was the
# oldest to support WAL mode, and 3.7.11 added multiple VALUE syntax.
# Prior to 3.7.6, you can't use LENGTH() on a BLOB column.
SQ3_IS_MIN_VERSION = SQ3_SUPPORTS_CTE

logger = log = logging.getLogger(__name__)

def _normalize_path(options):
    path = os.path.expanduser(os.path.expandvars(options.cache_local_dir))
    path = os.path.abspath(path)
    return path

def trace_file(options, prefix):
    # Return an open file for tracing to, if that is set up.
    # Otherwise, return nothing.

    # We choose a trace file based on ZEO_CACHE_TRACE. If it is
    # set to 'single', then we use a single file (not suitable for multiple
    # process, but records client opens/closes). If it is set to any other value,
    # we include a pid. If it is not set, we do nothing.
    trace = os.environ.get("ZEO_CACHE_TRACE")
    if not trace or not options.cache_local_dir:
        return None

    if trace == 'single':
        pid = 0
    else: # pragma: no cover
        pid = os.getpid()

    name = 'relstorage-trace-' + prefix + '.' + str(pid) + '.trace'

    parent_dir = _normalize_path(options)
    try:
        os.makedirs(parent_dir)
    except os.error:
        pass
    fname = os.path.join(parent_dir, name)
    try:
        tf = open(fname, 'ab')
    except IOError as e: # pragma: no cover
        log.warning("Cannot write tracefile %r (%s)", fname, e)
        tf = None
    else:
        log.info("opened tracefile %r", fname)
    return tf



def sqlite_files(options, prefix):
    """
    Calculate the sqlite filename and return it, plus a function that will
    destroy the sqlite file.
    """
    parent_dir = getattr(options, 'cache_local_dir', options)
    # Allow for memory and temporary databases (empty string):
    if parent_dir != ':memory:' and parent_dir:
        parent_dir = _normalize_path(options)
        try:
            # make it if needed. try to avoid a time-of-use/check
            # race (not that it matters here)
            os.makedirs(parent_dir)
        except os.error:
            pass

        fname = os.path.join(parent_dir, 'relstorage-cache2-' + prefix + '.sqlite3')
        wal_fname = fname + '-wal'
        shm_fname = fname + '-shm'
        def real_destroy():
            logger.info("Replacing any existing cache at %s", fname)
            __quiet_remove(fname)
            __quiet_remove(wal_fname)
            __quiet_remove(shm_fname)
        destroy = real_destroy
    else:
        fname = parent_dir
        wal_fname = None
        def noop_destroy():
            "Nothing to do."
        destroy = noop_destroy
    return fname, destroy


class Sqlite3TooOldError(ValueError):
    """Raised if the sqlite3 module is too old."""

    def __init__(self, *_args):
        super(Sqlite3TooOldError, self).__init__(
            "Unable to use sqlite; minimum version is %s but this version is %s" % (
                "3.8.3", sqlite3.sqlite_version_info
            ))


class _Driver(Sqlite3Driver):
    STATIC_AVAILABLE = SQ3_SUPPORTS_CTE
    DriverNotAvailableError = Sqlite3TooOldError

CORRUPT_DB_EXCEPTIONS = (sqlite3.DatabaseError,)
FAILURE_TO_OPEN_DB_EXCEPTIONS = (sqlite3.OperationalError, Sqlite3TooOldError)


def sqlite_connect(options, prefix,
                   overwrite=False,
                   **connect_kwargs):
    """
    Return a DB-API Connection object.

    If the database is corrupted, the file will be removed and recreated.

    .. caution:: Using the connection as a context manager does **not**
       result in the connection being closed, only committed or rolled back.
    """

    if not SQ3_IS_MIN_VERSION: # pragma: no cover
        raise Sqlite3TooOldError()

    fname, destroy = sqlite_files(options, prefix)

    corrupt_db_ex = CORRUPT_DB_EXCEPTIONS
    if overwrite:
        destroy()
        corrupt_db_ex = ()

    connect_args = (fname,)

    try:
        connection = _Driver().connect_to_file(
            *connect_args,
            **connect_kwargs
        )
    except corrupt_db_ex as e:
        __traceback_info__ = e, fname, destroy
        logger.exception("Corrupt cache database at %s; replacing", fname)
        destroy()
        connection = _Driver().connect_to_file(
            *connect_args,
            **connect_kwargs
        )

    return connection


def __quiet_remove(path):
    try:
        os.unlink(path)
    except os.error as e:
        # TODO: Use FileNotFoundError on Python 3
        log_meth = log.exception
        if e.errno == errno.ENOENT:
            log_meth = log.debug
        log_meth("Failed to remove %r", path)
        return False
    else:
        return True



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostgreSQL adapter for RelStorage."""
from __future__ import absolute_import, print_function

import logging
import re

from zope.interface import implementer

from ..adapter import AbstractAdapter

from ..dbiter import HistoryFreeDatabaseIterator
from ..dbiter import HistoryPreservingDatabaseIterator
from ..interfaces import IRelStorageAdapter
from ..packundo import HistoryFreePackUndo
from ..packundo import HistoryPreservingPackUndo
from ..poller import Poller
from ..schema import Schema
from ..scriptrunner import ScriptRunner
from . import drivers
from .batch import PostgreSQLRowBatcher
from .connmanager import Psycopg2ConnectionManager
from .locker import PostgreSQLLocker
from .mover import PostgreSQLObjectMover

from .oidallocator import PostgreSQLOIDAllocator
from .schema import PostgreSQLSchemaInstaller
from .stats import PostgreSQLStats
from .txncontrol import PostgreSQLTransactionControl


log = logging.getLogger(__name__)


# TODO: Move to own file
class PGPoller(Poller):

    _poll_newest_tid_query = Schema.all_transaction.select(
        Schema.all_transaction.c.tid
    ).order_by(
        Schema.all_transaction.c.tid, dir='DESC'
    ).limit(
        1
    ).prepared()


@implementer(IRelStorageAdapter)
class PostgreSQLAdapter(AbstractAdapter):
    """PostgreSQL adapter for RelStorage."""
    # pylint:disable=too-many-instance-attributes

    driver_options = drivers

    def __init__(self, dsn='', options=None, oidallocator=None,
                 locker=None,
                 mover=None,
                 connmanager=None
                 ):
        # options is a relstorage.options.Options or None
        self._dsn = dsn
        self.oidallocator = oidallocator
        self.locker = locker
        self.mover = mover
        self.connmanager = connmanager
        super(PostgreSQLAdapter, self).__init__(options)

    def _create(self):
        driver = self.driver
        options = self.options
        dsn = self._dsn

        self.version_detector = PostgreSQLVersionDetector()
        if self.connmanager is None:
            self.connmanager = Psycopg2ConnectionManager(
                driver,
                dsn=dsn,
                options=options,
            )
        self.runner = ScriptRunner()
        if self.locker is None:
            self.locker = PostgreSQLLocker(
                options,
                driver,
                PostgreSQLRowBatcher,
            )
        self.schema = PostgreSQLSchemaInstaller(
            options=options,
            connmanager=self.connmanager,
            runner=self.runner,
            locker=self.locker,
        )

        if self.mover is None:
            self.mover = PostgreSQLObjectMover(
                driver,
                options=options,
                runner=self.runner,
                version_detector=self.version_detector,
                batcher_factory=PostgreSQLRowBatcher,
            )
        if self.oidallocator is None:
            self.oidallocator = PostgreSQLOIDAllocator()

        self.poller = PGPoller(
            self.driver,
            keep_history=self.keep_history,
            runner=self.runner,
            revert_when_stale=options.revert_when_stale,
            transactions_may_go_backwards=(
                self.connmanager.replica_selector is not None
                or self.connmanager.ro_replica_selector is not None
            )
        )

        self.txncontrol = PostgreSQLTransactionControl(
            connmanager=self.connmanager,
            poller=self.poller,
            keep_history=self.keep_history,
            Binary=driver.Binary,
        )

        if self.keep_history:
            self.packundo = HistoryPreservingPackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            self.dbiter = HistoryPreservingDatabaseIterator(
                driver,
            )
        else:
            self.packundo = HistoryFreePackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            # TODO: Subclass for this.
            self.packundo._lock_for_share = 'FOR KEY SHARE OF object_state'
            self.dbiter = HistoryFreeDatabaseIterator(
                driver,
            )

        self.stats = PostgreSQLStats(
            connmanager=self.connmanager,
            keep_history=self.keep_history
        )

    def new_instance(self):
        inst = type(self)(
            dsn=self._dsn,
            options=self.options,
            oidallocator=self.oidallocator.new_instance(),
            locker=self.locker,
            mover=self.mover,
            connmanager=self.connmanager,
        )
        return inst

    def __str__(self):
        parts = []
        if self.keep_history:
            parts.append('history preserving')
        else:
            parts.append('history free')
        dsnparts = self._dsn.split()
        s = ' '.join(p for p in dsnparts if not p.startswith('password'))
        parts.append('dsn=%r' % s)
        return "<%s at %x %s>" % (
            self.__class__.__name__, id(self), ",".join(parts)
        )

    __repr__ = __str__

    # A temporary magic variable as we move TID allocation into some
    # databases; with an external clock, we *do* need to sleep waiting for
    # TIDs to change in a manner we can exploit; that or we need to be very
    # careful about choosing pack times.
    RS_TEST_TXN_PACK_NEEDS_SLEEP = 1

    def lock_database_and_choose_next_tid(self,
                                          cursor,
                                          username,
                                          description,
                                          extension):
        proc_name = 'SELECT lock_and_choose_tid'
        proc = proc_name + '()'
        args = ()
        if self.keep_history:
            # (packed, username, descr, extension)
            proc = proc_name + '(%s, %s, %s, %s)'
            b = self._binary
            args = (False, b(username), b(description), b(extension))

        cursor.execute(proc, args)
        tid, = cursor.fetchone()
        return tid

    def lock_database_and_move(self,
                               store_connection, load_connection,
                               transaction_has_blobs, # pylint:disable=unused-argument
                               ude,
                               commit=True,
                               committing_tid_int=None,
                               after_selecting_tid=lambda tid: None):

        # In all versions of Postgres (up through 11 anyway),
        # stored functions cannot COMMIT. In Postgres 11,
        # the newly-introduced stored procedures *can* COMMIT,
        # if they're at the top level; that includes anonymous
        # DO blocks, BUT (and this goes for both anonymous and CALL'd procs)
        # ONLY if they're not already part of a transaction.
        #
        # Options:
        #
        # We can tack ``; COMMIT`` on to the end of the ``SELECT``
        # statement, but pg8000 doesn't like that ("cannot insert
        # multiple commands into a prepared statement") psycopg2 will
        # allow it, but because the last statement wasn't a ``SELECT``
        # we lose access to the TID.
        #
        # If we alter the temp tables to preserve their rows on
        # COMMIT, we could COMMIT now, turn on autocommit, and call
        # the function to move rows and make current. The problem
        # there is that we would lose our row locks, so we're not
        # guaranteed that we'd actually be able to finish the COMMIT.
        #
        # We can use the GUC (grand unified config) as session variables
        # and store the return value in the session (as text) and select it back out
        # after the commit. This seems to work, at the expense of extra
        # DB communication, but it gets the COMMIT to happen in one
        # trip to the DB: This is confirmed by database statement logging.
        # The only problem here is that it still fails on pg8000;
        # we'll just ignore that.
        params = (committing_tid_int, commit)
        # (p_committing_tid, p_commit)
        proc = 'lock_and_choose_tid_and_move(%s, %s)'
        if self.keep_history:
            username, description, extension = ude
            b = self._binary
            params += (b(username), b(description), b(extension))
            # (p_committing_tid, p_commit, p_user, p_desc, p_ext)
            proc = 'lock_and_choose_tid_and_move(%s, %s, %s, %s, %s)'

        cursor = store_connection.cursor
        if commit and self.driver.supports_multiple_statement_execute:
            # Do this all in one trip to the database so that we don't need to
            # wake up to handle the commit. Unfortunately, though, this
            # will make the internal state of our connection object in libpq not match
            # the actual state of the transaction on the server so we must still
            # execute connection.commit() to bring them back in sync. This results
            # in a warning on the server about no transaction being in progress.
            proc = (
                "SELECT SET_CONFIG('rs.tid', " + proc + "::text, FALSE); "
                "COMMIT; "
                "SELECT current_setting('rs.tid')"
            )
            self.driver.execute_multiple_statement_with_hidden_commit(
                store_connection.connection,
                store_connection.cursor,
                proc,
                params
            )
        else:
            proc = 'SELECT ' + proc
            cursor.execute(proc, params)

        tid_int, = cursor.fetchone()
        tid_int = int(tid_int)
        if commit:
            if self.driver.supports_multiple_statement_execute:
                self.driver.sync_status_after_hidden_commit(store_connection.connection)
            else:
                self.txncontrol.commit_phase2(store_connection, "-", load_connection)
        after_selecting_tid(tid_int)
        return tid_int, "-"


    DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = False

    def _best_lock_objects_and_detect_conflicts(self, cursor, read_current_oids):
        read_current_oids_p = None
        read_current_tids_p = None
        if read_current_oids:
            # Pass both the OIDs and TIDS and make the database do an extra query
            # to filter the non-matching so that we only have to deal with rows that
            # actually conflict. This keeps the Python-level loop that we do once the rows are
            # locked as short as possible.
            # Separate arrays are faster to deal with than an array of tuples,
            # they just have to be in the same order.

            # Just pass in the OIDs and let it return to us the committed tids
            # which we will loop over to compare. This simplifies the SQL
            # and reduces the number of queries we have to do. If that looping is
            # a problem
            read_current_oids_p = []
            read_current_tids_p = []
            for k, v in read_current_oids.items():
                read_current_oids_p.append(k)
                read_current_tids_p.append(v)

        cursor.execute('SELECT * FROM lock_objects_and_detect_conflicts(%s, %s)',
                       (read_current_oids_p, read_current_tids_p,))
        conflicts = cursor.fetchall()
        return conflicts

    def _describe_best_lock_objects_and_detect_conflicts(self):
        return 'lock_objects_and_detect_conflicts(%s)'


class PostgreSQLVersionDetector(object):

    version = None

    def get_version(self, cursor):
        """Return the (major, minor) version of the database"""
        if self.version is None:
            cursor.execute("SELECT version()")
            v = cursor.fetchone()[0]
            m = re.search(r"([0-9]+)[.]([0-9]+)", v)
            if m is None:
                raise AssertionError("Unable to detect database version: " + v)
            self.version = int(m.group(1)), int(m.group(2))
        return self.version



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Test mixin dealing with packing, especially concurrently.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools


from persistent.mapping import PersistentMapping
import transaction
from ZODB.DB import DB
from ZODB.serialize import referencesf

from .reltestbase import RelStorageTestBase


class TestPackBase(RelStorageTestBase):
    # pylint:disable=abstract-method

    def setUp(self):
        super(TestPackBase, self).setUp()
        self.main_db = self._closing(DB(self._storage))

    def tearDown(self):
        self.main_db.close()
        self.main_db = None
        super(TestPackBase, self).tearDown()

    OID_ROOT = 0
    OID_A = 1
    OID_B = 2
    OID_C = 3
    OID_D = 4

    OID_INITIAL_SET = (OID_ROOT, OID_A, OID_B, OID_C)

    def _create_initial_state(self):
        # Given a set of referencing objects present at the beginning
        # of the pre pack:
        #          0    1    2    3
        #   T1: root -> A -> B -> C
        #
        # If a new transaction is committed such that the graph becomes:
        #
        #         0    1
        #  T2: root -> A
        #          \-> B -> D -> C
        #              2    4    3
        #
        # That is, C is no longer referenced from B but a new object
        # D, B is referenced not from A but from the root.
        txm = transaction.TransactionManager(explicit=True)
        conn = self.main_db.open(txm)
        txm.begin()

        A = conn.root.A = PersistentMapping() # OID 0x1
        B = A['B'] = PersistentMapping() # OID 0x2
        C = B['C'] = PersistentMapping() # OID 0x3

        txm.commit()
        oids = {
            'A': A._p_oid,
            'B': B._p_oid,
            'C': C._p_oid,
        }
        conn.close()

        return oids

    def _mutate_state(self, initial_oids, *_args, **_kwargs):
        txm = transaction.TransactionManager(explicit=True)
        conn = self.main_db.open(txm)
        txm.begin()

        A = conn.root.A
        B = A['B']
        del A['B']
        D = B['D'] = PersistentMapping() # OID 0x4
        C = B['C']
        del B['C']
        D['C'] = C

        txm.commit()

        initial_oids['D'] = D._p_oid
        conn.close()

    def _check_inject_changes(
            self,
            initial_oids, # type: Dict[str, bytes]
            inject_changes,
            delta=1,
            hook='on_filling_object_refs_added'
    ):
        # We expect inject_changes to mutate *initial_oids* by *delta*
        # (see _mutate_state)
        len_initial_oids = len(initial_oids)

        adapter = self._storage._adapter
        setattr(adapter.packundo, hook, inject_changes)

        # Pack to the current TID (RelStorage extension)
        packtime = None
        self._storage.pack(packtime, referencesf)

        # The on_filling_object_refs_added hook should have been called once
        self.assertEqual(len(initial_oids),
                         len_initial_oids + delta,
                         initial_oids)

        # All children should still exist.
        missing = {}
        for name, oid in initial_oids.items():
            try:
                state, _tid = self._storage.load(oid, '')
            except KeyError as e:
                missing[name] = e
            else:
                self.assertIsNotNone(state)

        self.assertEmpty(missing)

    def test_pack_when_object_ref_moved_during_prepack(self):
        # If we mutate after we gather the initial list of
        # objects but before we start examining their state
        # we shouldn't collect any objects.
        expect_oids = self._create_initial_state()

        inject_changes = functools.partial(
            self._mutate_state,
            expect_oids
        )

        self._check_inject_changes(expect_oids,
                                   inject_changes)

    # https://ci.appveyor.com/project/jamadden/relstorage/build/1.0.19/job/a1vq619n84ss1s9a
    def test_pack_when_referring_object_mutates_during_prepack(self):
        # Packing should not remove objects referenced by an object
        # that changes during packing. In this case, we're adding a
        # new object (which should be untouched because it's not
        # visible as-of the pack-TID) and mutating an existing object
        # to refer to the new object as well as the new object. add
        # some data to be packed

        expect_oids = self._create_initial_state()

        def inject_changes(**_kwargs):
            # Change the database just after the list of objects
            # to analyze has been determined.
            txm = transaction.TransactionManager(explicit=True)
            conn = self.main_db.open(txm)
            txm.begin()
            Added = conn.root.ADDED = PersistentMapping()
            txm.commit()
            expect_oids['Added'] = Added._p_oid
            conn.close()

        self._check_inject_changes(expect_oids,
                                   inject_changes)

class HistoryFreeTestPack(TestPackBase):
    # pylint:disable=abstract-method
    keep_history = False

    def test_pack_when_object_ref_moved_after_ref_finding_first_batch(self):
        # If we mutate after gather the initial list of objects, and after
        # finding references in the first batch (of all objects), we should not
        # collect anything.
        #
        # Even though this would seem to be similar to the 'ref_moved_during_prepack' case,
        # it is different. This currently fails because even though we find a reference
        # to the object, it's from a state that doesn't exist anymore
        # and we don't bother to look at it.
        expect_oids = self._create_initial_state()

        def inject_changes(oid_batch, refs_found):
            # We're examining the root and the first three objects
            self.assertEqual(sorted(oid_batch), sorted(self.OID_INITIAL_SET))
            self.assertEqual(len(refs_found), 3)
            # We're only called once: a single batch
            self.assertNotIn('D', expect_oids)
            self._mutate_state(expect_oids)

        self._check_inject_changes(
            expect_oids,
            inject_changes,
            hook='on_fill_object_ref_batch',
        )

        self.assertIn('D', expect_oids) # hook was called.

    def test_pack_when_object_ref_moved_just_before_examine_prev_reference(self):
        # We mutate just before we examine the state for the B object.
        expect_oids = self._create_initial_state()
        seen_oids = []
        def inject_changes(oid_batch, refs_found): # pylint:disable=unused-argument
            oid_batch = list(oid_batch)
            seen_oids.extend(oid_batch)
            self.assertEqual(1, len(oid_batch))
            if oid_batch == [self.OID_A]:
                # Next batch will be B.
                self._mutate_state(expect_oids)

        # Limit batch size and force commits.
        packundo = self._storage._adapter.packundo
        packundo.fill_object_refs_batch_size = 1
        packundo.fill_object_refs_commit_frequency = 0

        try:
            self._check_inject_changes(
                expect_oids,
                inject_changes,
                hook='on_fill_object_ref_batch',
            )
        finally:
            self.assertEqual(sorted(seen_oids), sorted(self.OID_INITIAL_SET))
            self.assertIn('D', expect_oids) # hook was called.


class HistoryPreservingTestPack(TestPackBase):
    # pylint:disable=abstract-method
    keep_history = True



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
import os.path
import sqlite3
import functools

from zope.interface import implementer

from relstorage._compat import PY3
from relstorage._compat import PY2
from relstorage._compat import PY36
from relstorage._util import log_timed

from ..drivers import implement_db_driver_options
from ..drivers import AbstractModuleDriver
from ..drivers import GeventDriverMixin
from ..drivers import MemoryViewBlobDriverMixin
from ..interfaces import IDBDriver
from ..interfaces import IDBDriverSupportsCritical
from .._util import DatabaseHelpersMixin

from .dialect import Sqlite3Dialect

__all__ = [
    'Sqlite3Driver',
    'Sqlite3GeventDriver',
]

database_type = 'sqlite3'
logger = __import__('logging').getLogger(__name__)



_MB = 1024 * 1024
DEFAULT_MAX_WAL = 10 * _MB

# Benchmarking on at least one system doesn't show an improvement to
# either reading or writing by forcing a large mmap_size (maybe that's
# not enabled on macOS?). By default, just report. A setting of 0
# means do not use.
DEFAULT_MMAP_SIZE = None
# 4096 is the page size in current releases of sqlite; older versions
# used 1024. A larger page makes sense as we have biggish values.
# Going larger doesn't make much difference in benchmarks.
DEFAULT_PAGE_SIZE = 4096
# The default of -2000 is abs(N*1024) = ~2MB. At 4K page size,
# that's about 500 pages. A negative number specifies bytes, a
# positive number specifies pages. In general, since we have two
# connections for each ZODB connection, we're probably best letting
# the cache stay relatively small and letting the operating system
# page cache decide between threads and process.s
DEFAULT_CACHE_SIZE = None
# Control where temporary data is:
#
# FILE = a deleted disk file (that sqlite never flushes so
# theoretically just exists in the operating system's filesystem
# cache)
#
# MEMORY = explicitly in memory only
#
# DEFAULT = compile time default. Benchmarking for large writes
# doesn't show much difference between FILE and MEMORY *usually*:
# Sometimes changing the setting back and forth causes massive
# changes in performance, at least on a macOS desktop system.
DEFAULT_TEMP_STORE = None

# How long before we get 'OperationalError: database is locked'
DEFAULT_TIMEOUT = 15


class UnableToConnect(sqlite3.OperationalError):
    filename = None
    def with_filename(self, f):
        self.filename = f
        return self

    def __str__(self):
        s = super(UnableToConnect, self).__str__()
        if self.filename:
            s += " (At: %r)" % self.filename
        return s


class Cursor(sqlite3.Cursor):
    has_analyzed_temp = False

    def execute(self, stmt, params=None):
        # While we transition away from hardcoded SQL to the query
        # objects, we still have some %s params out there. This papers
        # over that.
        if params is not None:
            stmt = stmt.replace('%s', '?')
            return sqlite3.Cursor.execute(self, stmt, params)

        return sqlite3.Cursor.execute(self, stmt)

    def executemany(self, stmt, params):
        stmt = stmt.replace('%s', '?')
        return sqlite3.Cursor.executemany(self, stmt, params)

    # PyPy likes to throw `sqlite3.ProgrammingError: Cannot operate on a closed cursor`
    # under various circumstances. Which ones depend on which version of PyPy.
    # For example, upgrading from PyPy3.6 7.1 to PyPy3.6 7.3.1 started to do
    # this in ``self.connection`` when printing ``__traceback_info__``
    # on failures (""):
    #
    #  File "/opt/python/pypy3.6-7.3.1/lib_pypy/_sqlite3.py", line 884, in __execute
    #     if self.__connection._begin_statement and self.__statement._is_dml:
    #  AttributeError: 'Connection' object has no attribute '_begin_statement'
    #

    @property
    def __rs_connection(self):
        try:
            return self.connection
        except sqlite3.ProgrammingError: # pragma: no cover
            return "<closed cursor>"

    def __repr__(self):
        return '<%s at 0x%x from %r>' % (
            type(self).__name__,
            id(self),
            self.__rs_connection
        )

    def close(self):
        try:
            sqlite3.Cursor.close(self)
        except sqlite3.ProgrammingError: # pragma: no cover
            # XXX: Where was this happening? Why are we doing this?
            # (I haven't seen it locally)
            pass


class _ExplainCursor(DatabaseHelpersMixin): # pragma: no cover (A debugging aid)
    def __init__(self, cur):
        self.cur = cur

    def __getattr__(self, name):
        return getattr(self.cur, name)

    def __iter__(self):
        return iter(self.cur)

    def execute(self, sql, *args):
        if sql.strip().startswith(('INSERT', 'SELECT', 'DELETE', 'WITH')):
            exp = 'EXPLAIN QUERY PLAN ' + sql.lstrip()
            print(sql)
            self.cur.execute(exp, *args)
            print(self._rows_as_pretty_string(self.cur))
        return self.cur.execute(sql, *args)


class Connection(sqlite3.Connection):
    CURSOR_FACTORY = Cursor
    has_analyzed_temp = False
    before_commit_functions = ()
    _rs_has_closed = False
    _rs_progress_handler = None
    replica = None
    standard_progress_handler = (None, 0)
    # For PyPy3.6-7.3.1. We connect with isolation_level=None, leaving
    # sqlite in its own transaction management mode. But PyPy
    # 3.6-7.3.1 has that broken prior to
    # https://foss.heptapod.net/pypy/pypy/-/commit/fbbe06715eb48df1a03640672d99335695d3e47c.
    # See also https://foss.heptapod.net/pypy/pypy/-/issues/3226
    _begin_statement = None

    def __init__(self, rs_db_filename, *args, **kwargs):
        __traceback_info__ = args, kwargs
        # PyPy3 calls self.commit() during __init__.
        self.rs_db_filename = rs_db_filename
        self.before_commit_functions = []

        try:
            super(Connection, self).__init__(*args, **kwargs)
        except sqlite3.OperationalError as e:
            raise UnableToConnect(e).with_filename(rs_db_filename)

    def __repr__(self):
        if not self.rs_db_filename:
            return super(Connection, self).__repr__()
        try:
            in_tx = self.in_transaction
        except sqlite3.ProgrammingError:
            in_tx = 'closed'

        return '<%s at 0x%x to %r in_transaction=%s>' % (
            type(self).__name__,
            id(self), self.rs_db_filename,
            in_tx
        )

    def _at_transaction_end(self):
        pass

    if not PY3:
        # This helpful attribute is Python 3 only.
        assert not hasattr(sqlite3.Connection, 'in_transaction')
        __in_transaction = None
        @property
        def in_transaction(self):
            return self.__in_transaction

        def _at_transaction_end(self):
            self.__in_transaction = None

    def register_before_commit_cleanup(self, func):
        self.before_commit_functions.append(func)

    if 0: # pylint:disable=using-constant-test
        def cursor(self):
            return _ExplainCursor(sqlite3.Connection.cursor(self, self.CURSOR_FACTORY))
    else:
        def cursor(self):
            return sqlite3.Connection.cursor(self, self.CURSOR_FACTORY)

    def return_to_repeatable_read(self):
        # See mover.py
        # deliberately doesn't use self.commit(), we don't want to run cleanup
        # hooks, this is a mid-logical-ZODB-transaction operation
        sqlite3.Connection.commit(self)

    def commit(self):
        try:
            for func in self.before_commit_functions:
                func(self)
            sqlite3.Connection.commit(self)
        finally:
            self._at_transaction_end()


    def rollback(self):
        try:
            sqlite3.Connection.rollback(self)
            for func in self.before_commit_functions:
                func(self, True)
        finally:
            self._at_transaction_end()

    def __check_and_log(self, reported_values, changed_values):
        logger.debug(
            "Connection: %(conn)s.\n\t"
            "Using sqlite3 version: %(ver)s.\n\t"
            "Default    connection settings: %(defs)s.\n\t"
            "Changing   connection settings: %(changing)s.\n\t"
            "Desired    connection settings: %(desired)s.\n\t"
            "Unapplied  connection settings: %(applied)s.\n\t",
            dict(
                conn=self,
                ver=sqlite3.sqlite_version,
                defs={
                    k: v for k, v in reported_values.items()
                    if k not in changed_values
                },
                changing={
                    k: v for k, v in reported_values.items()
                    if k in changed_values
                },
                desired={
                    k: v[0] for k, v in changed_values.items()
                },
                applied={
                    k: v[1] for k, v in changed_values.items()
                    if v[0] != v[1]
                },
            )
        )

    # PRAGMA statements don't allow ? placeholders
    # when executed. This is probably a bug in the sqlite3
    # module.
    def __execute_pragma(self, name, value):
        stmt = 'PRAGMA %s = %s' % (name, value)
        cur = self.execute(stmt)
        # On PyPy, it's important to traverse the cursor, even if
        # you don't expect any results, because it still counts as
        # a statement that's open and can cause 'OperationalError:
        # can't commit with SQL operations active'.
        return cur.fetchall()

    def execute_pragmas(self, **kwargs):
        report = {}
        changed = {} # {setting: (desired, updated)}
        if 'page_size' in kwargs:
            # This has to happen before changing into WAL.
            ps = kwargs.pop('page_size')
            items = [('page_size', ps)]
            items.extend(kwargs.items())
        else:
            items = sorted(kwargs.items())

        for pragma, desired in items:
            # Query, report, then change
            stmt = 'PRAGMA %s' % (pragma,)
            row, = self.execute(stmt).fetchall() or ((),)
            orig_value = row[0] if row else None
            assert pragma not in report # Only one
            report[pragma] = orig_value
            if desired is not None and desired != orig_value:
                self.__execute_pragma(pragma, desired)
                row = self.execute(stmt).fetchone()
                new_value = row[0] if row else None
                changed[pragma] = (desired, new_value)

        self.__check_and_log(report, changed)

    @log_timed
    def close(self):
        # If we're the only connection open to this database,
        # and SQLITE_FCNTL_PERSIST_WAL is true (by default
        # *most* places, but apparently not in the sqlite3
        # 3.24 shipped with Apple in macOS 10.14.5), then when
        # we close the database the wal file that was built up
        # by any of the writes that have been done will be automatically
        # combined with the database file, as if with
        # "PRAGMA wal_checkpoint(RESTART)".
        #
        # This can be slow. It releases the GIL, so it could be done in another thread;
        # but most often (aside from tests) we're closing connections because we're
        # shutting down the process so spawning threads isn't really a good thing.
        if self._rs_has_closed: # pragma: no cover
            return

        self._rs_has_closed = True
        self.before_commit_functions = ()
        # Recommended best practice is to OPTIMIZE the database for
        # each closed connection. OPTIMIZE needs to run in each connection
        # so it can see what tables and indexes were used. It's usually fast,
        # but has the potential to be slow and lock the database. It cannot be executed
        # on a read-only connection. We don't want to block too long just closing
        # a connection, so reset the time we wait to get a lock
        # (if needed) to a lower value (ms).
        try:
            self.executescript("""
            PRAGMA busy_timeout = 3000;
            PRAGMA query_only = 0;
            PRAGMA optimize;
            """)
        except sqlite3.OperationalError:
            logger.debug("Failed to optimize databas, probably in use", exc_info=True)
        except sqlite3.DatabaseError:
            # It's possible the file was removed.
            logger.exception("Failed to optimize database; was it removed?")

        super(Connection, self).close()


@implementer(IDBDriver)
class Sqlite3Driver(MemoryViewBlobDriverMixin,
                    AbstractModuleDriver):
    dialect = Sqlite3Dialect()
    __name__ = 'sqlite3'
    MODULE_NAME = __name__
    STATIC_AVAILABLE = (
        # Tested to work on Python 2.7 and Python 3.6+;
        # seen some strange ``ProgrammingError: closed``
        # on Python 3.5
        (PY2 or PY36)
        # 3.11 is the oldest version tested on CI
        and sqlite3.sqlite_version_info[:2] >= (3, 11)
    )

    CONNECTION_FACTORY = Connection
    DEFAULT_CONNECT_ARGS = {
    }
    supports_64bit_unsigned_id = False

    def __init__(self):
        super(Sqlite3Driver, self).__init__()
        # Sadly, if a connection is closed out from under it,
        # sqlite3 throws ProgrammingError, which is not very helpful.
        # That should really only happen in tests, though, since
        # we're directly connected to the file on disk.
        self.disconnected_exceptions += (self.driver_module.ProgrammingError,) # pylint:disable=no-member
        # Make our usual connect() method call our connect_to_file method instead of the
        # module's connect() method so we get our preferred goodies.
        self._connect = self.connect_to_file

    if PY3:
        # in_transaction doesn't work on Py2, so assume the worst
        # by inheriting the functions.
        def connection_may_need_rollback(self, conn):
            try:
                return conn.in_transaction
            except sqlite3.ProgrammingError:
                # we're closed. We do need to attempt the rollback so
                # we catch the error and know to drop the connection.
                return True

        connection_may_need_commit = connection_may_need_rollback

        # Py2 returns buffer for blobs, hence the Mixin.
        # But Py3 returns plain bytes.
        def binary_column_as_state_type(self, data):
            return data



    def _connect_to_file_or_uri(self, fname,
                                timeout=DEFAULT_TIMEOUT,
                                pragmas=None,
                                quick_check=True,
                                isolation_level=None,
                                factory_args=(),
                                **connect_args):

        factory_args = (fname,) + factory_args
        factory = lambda *args, **kwargs: self.CONNECTION_FACTORY(*(factory_args + args), **kwargs)

        default_connect_args = self.DEFAULT_CONNECT_ARGS.copy()
        default_connect_args.update(connect_args)
        connect_args = default_connect_args

        connection = sqlite3.connect(
            fname,
            # See Sqlite3ConnectionManager for the meaning of this.
            # We default to None which puts us in the underlying sqlite autocommit
            # mode. This is unlike the standard library, so you must
            # either set this or manage your own transactions explicitly
            isolation_level=isolation_level,
            factory=factory,
            # We explicitly push closing off to a new thread.
            check_same_thread=False,
            timeout=timeout,
            **connect_args
        )

        if str is bytes:
            # We don't use the TEXT type, but even so
            # sqlite complains:
            #
            # ProgrammingError: You must not use 8-bit bytestrings unless
            # you use a text_factory that can interpret 8-bit bytestrings
            # (like text_factory = str). It is highly recommended that you
            # instead just switch your application to Unicode strings.
            connection.text_factory = str

        # Make sure we have at least one pragma that touches
        # the database so that we can verify that it's not corrupt.
        pragmas = pragmas or {}
        pragmas.setdefault('journal_mode', 'wal')
        cur = connection.cursor()
        __traceback_info__ = fname, cur, pragmas
        try:
            connection.execute_pragmas(**pragmas)
            # Quick integrity check before we read.
            if quick_check:
                cur.execute('PRAGMA quick_check')
                rows = cur.fetchall()
                if len(rows) != 1 or rows[0][0] != 'ok':
                    msg = '\n'.join(row[0] for row in rows)
                    raise sqlite3.DatabaseError('Quick integrity check failed %s' % msg)
        except:
            logger.exception("Failed to execute pragmas")
            cur.close()
            connection.close()
            raise

        cur.close()

        return connection

    def connect_to_file(self, fname,
                        query_only=False,
                        max_wal_size=DEFAULT_MAX_WAL,
                        mmap_size=DEFAULT_MMAP_SIZE,
                        page_size=DEFAULT_PAGE_SIZE,
                        cache_size=DEFAULT_CACHE_SIZE,
                        temp_store=DEFAULT_TEMP_STORE,
                        timeout=DEFAULT_TIMEOUT,
                        quick_check=True,
                        isolation_level=None,
                        extra_pragmas=None,
                        override_pragmas=None):
        """
        Return a DB-API Connection object.

        .. caution:: Using the connection as a context manager does **not**
           result in the connection being closed, only committed or rolled back.
        """

        # There are some things we might like to do, but we can't
        # unless we're operating on at least Python 3.4, where the URI
        # syntax is supported. It also turns out that they don't really work either:
        #
        # - Enable shared cache. (https://www.sqlite.org/sharedcache.html)
        # This doesn't work because of locking: all connections sharing the cache
        # lock at the same time.
        # - Use ?mode=ro for true query_only mode. But if the file doesn't
        # exist that fails, and we can't execute certain pragmas that way either.

        fname = os.path.abspath(fname) if fname and fname != ':memory:' else fname

        # Things we like but don't require.
        # Note that we use the primitive values for these things so that we can
        # detect when they get set for reporting purposes.
        pragmas = {
            'journal_size_limit': max_wal_size,
            'mmap_size': mmap_size,
            'page_size': page_size,
            'temp_store': temp_store,
            # WAL mode is always consistent even after a operating system
            # crash in NORMAL mode. It might lose a transaction, though.
            # The default is often FULL/2, which is higher than NORMAL/1.
            'synchronous': 1,
            # If cache_spill is allowed, at some point a transaction
            # can begin writing dirty pages to the database file, taking
            # an exclusive lock. That could be at arbitrary times, so we don't want that.
            'cache_spill': 0,

            # In the past, we considered disabling auto-checkpoint so that commits have
            # reliable duration; after commit, if it's a good time,
            # we can run 'PRAGMA wal_checkpoint'. (In most cases, the last
            # database connection that's open will essentially do that
            # automatically.) This is a balancing act, though, and
            # we expose this setting to the end user to allow them to tune it.
            # By default, use the compiled-in default and report what that is.
            'wal_autocheckpoint': None,
            # Things to query and report.
            'soft_heap_limit': None, # 0 means no limit
            'cache_size': cache_size,
            # How big is the database?
            'page_count': None,
            'busy_timeout': None,
            'query_only': None,
        }

        # User-specified extra pragmas go here.
        pragmas.update(extra_pragmas or {})

        # Things that *must* be set.
        required_pragmas = {
            # WAL mode can actually be a bit slower at commit time,
            # but buys us far better concurrency.
            # Note: In-memory databases always use 'memory' as the journal mode;
            # temporary databases always use 'delete'.
            'journal_mode': 'wal',
        }

        if query_only:
            required_pragmas['query_only'] = 1

        pragmas.update(required_pragmas)

        pragmas.update(override_pragmas or {})

        return self._connect_to_file_or_uri(
            fname,
            pragmas=pragmas,
            timeout=timeout,
            quick_check=quick_check,
            isolation_level=isolation_level
        )




###
# Gevent.
#
# Database drivers that use networking and a gevent cooperative driver
# will essentially always switch while any given query is running.
# This means that acquiring database locks doesn't block the Python
# process and other greenlets can run. We need to achieve the same
# thing here, otherwise we can deadlock the process --- despite the
# 'critical_phase', it's still possible that, while already holding
# sqlite's exclusive locks, we could switch to another greenlet that
# wants to take those same locks. If we block, we deadlock until a
# timeout occurs. The ZODB tests check7*Threads do this explicitly by
# invoking tpc_vote() and then time.sleep().
#
# So, since there are limited statements that we execute to lock,
# we push those onto gevent's threadpool (unless we're in critical phase).
#
# The rest of the time, we use a progress handler for non-blocking functions
# to periodically automatically switch. This is a bit non-deterministic,
# but low overhead.
###

def run_blocking_in_threadpool(func):
    @functools.wraps(func)
    def in_threadpool(self, stmt, params=None):
        if self.could_block(stmt) and not self.connection.is_in_critical_phase():
            # Be sure not to yield in the threadpool thread, there's nothing to yield to.
            with self.connection.temp_critical_phase():
                return self.connection.gevent.get_hub().threadpool.apply(
                    func, (self, stmt, params)
                )
        return func(self, stmt, params)
    return in_threadpool

class GeventCursor(Cursor):

    BLOCKING_PREFIXES = (
        'BEGIN EXCLUSIVE',
        'BEGIN IMMEDIATE',
        'UPDATE',
        'INSERT',
    )

    def could_block(self, stmt):
        stmt = stmt.upper().rstrip()
        if stmt.upper().rstrip().startswith(self.BLOCKING_PREFIXES):
            return True
        return False

    execute = run_blocking_in_threadpool(Cursor.execute)
    executemany = run_blocking_in_threadpool(Cursor.executemany)


class GeventConnection(Connection):
    CURSOR_FACTORY = GeventCursor

    _in_critical_phase = None

    def __init__(self, rs_db_filename, gevent, yield_interval,
                 *args, **kwargs):
        # PyPy3 calls commit() from early in __init__, and certain
        # things aren't set up yet. Notably, set_progress_handler
        # doesn't work.
        self._at_transaction_end = lambda: None
        self.yield_interval = yield_interval
        self.gevent = gevent
        Connection.__init__(self, rs_db_filename, *args, **kwargs)
        del self._at_transaction_end
        self.set_progress_handler(gevent.sleep, yield_interval)

    def is_in_critical_phase(self):
        return self._in_critical_phase

    def _at_transaction_end(self): # pylint:disable=method-hidden
        self.exit_critical_phase()

    def enter_critical_phase_until_transaction_end(self):
        self._in_critical_phase = True
        self.set_progress_handler(None, 0)

    def exit_critical_phase(self):
        self._in_critical_phase = None
        self.set_progress_handler(self.gevent.sleep, self.yield_interval)

    @contextlib.contextmanager
    def temp_critical_phase(self):
        if self._in_critical_phase:
            yield
        else:
            self.enter_critical_phase_until_transaction_end()
            try:
                yield
            finally:
                self.exit_critical_phase()

    # Note that we don't commit or rollback on the threadpool. If the
    # threadpool is full, we don't want to put an operation (releasing
    # locks) that is needed for threadpool tasks (taking locks) on the
    # back of the threadpool.


@implementer(IDBDriverSupportsCritical)
class Sqlite3GeventDriver(GeventDriverMixin,
                          Sqlite3Driver):

    __name__ = 'gevent ' + Sqlite3Driver.MODULE_NAME
    _GEVENT_CAPABLE = True
    _GEVENT_NEEDS_SOCKET_PATCH = False
    CONNECTION_FACTORY = GeventConnection

    DEFAULT_CONNECT_ARGS = {
        # Keep a large number of cached statements. Certain
        # data is tracked at the cached statement level. In our
        # case that's notably the number of virtual machine instructions
        # it has executed.
        'cached_statements': 1024
    }

    # Call ``gevent.sleep()`` after this many virtual instructions
    # have been executed by the SQLite virtual machine for any one
    # statement (because instruction execution is tracked at the statement
    # level, not the database level).
    yield_to_gevent_instruction_interval = 100

    def configure_from_options(self, options):
        if options.adapter and options.adapter.config.gevent_yield_interval:
            conf = options.adapter.config
            self.yield_to_gevent_instruction_interval = conf.gevent_yield_interval

    def _connect_to_file_or_uri(self, *args, **kwargs): # pylint:disable=arguments-differ,signature-differs
        assert 'factory_args' not in kwargs
        kwargs['factory_args'] = (
            self.gevent,
            self.yield_to_gevent_instruction_interval,
        )
        return super(Sqlite3GeventDriver, self)._connect_to_file_or_uri(*args, **kwargs)

    def enter_critical_phase_until_transaction_end(self, connection, cursor): # pylint:disable=unused-argument
        connection.enter_critical_phase_until_transaction_end()

    def is_in_critical_phase(self, connection, cursor): # pylint:disable=unused-argument
        return connection.is_in_critical_phase()

    def exit_critical_phase(self, connection, cursor): # pylint:disable=unused-argument
        connection.exit_critical_phase()


implement_db_driver_options(
    __name__,
    '.drivers'
)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import, print_function

import logging

from zope.interface import implementer

from .interfaces import IPoller
from .interfaces import StaleConnectionError

from .schema import Schema
from .sql import func

log = logging.getLogger(__name__)

@implementer(IPoller)
class Poller(object):
    """Database change notification poller"""

    # The zoid is the primary key on both ``current_object`` (history
    # preserving) and ``object_state`` (history free), so these
    # queries are guaranteed to only produce an OID once.

    _poll_inv_query = Schema.all_current_object.select(
        Schema.all_current_object.c.zoid, Schema.all_current_object.c.tid
    ).where(
        Schema.all_current_object.c.tid > Schema.all_current_object.bindparam('tid')
    ).order_by(
        Schema.all_current_object.c.tid, 'DESC'
    ).prepared()

    _poll_newest_tid_query = Schema.all_transaction.select(
        func.max(Schema.all_transaction.c.tid)
    ).prepared()

    def __init__(self, driver, keep_history, runner,
                 revert_when_stale, transactions_may_go_backwards):
        self.driver = driver
        self.keep_history = keep_history
        self.runner = runner
        self.revert_when_stale = revert_when_stale
        self.transactions_may_go_backwards = transactions_may_go_backwards

    def get_current_tid(self, cursor):
        self._poll_newest_tid_query.execute(cursor)
        rows = cursor.fetchall() or ((0,),)
        current_tid, = rows[0]
        return current_tid or 0

    def poll_invalidations(self, conn, cursor, prev_polled_tid):
        """
        See ``IPoller``
        """
        # pylint:disable=unused-argument

        # Some databases, in some isolation modes, only establish a snapshot
        # of a particular table when the table is first accessed in a given transaction.
        # (Looking at you, MySQL on Windows). Thus if we're accessing two tables,
        # as we would in history-preserving mode, we could get slightly different answers:
        # The current_object table might move forward by a transaction while we're accessing the
        # transaction table, leading to the results being inconsistent.
        # For this reason, we only perform a single poll query against the actual object data.
        # We order this to get the newest TID first, and we return a chain iterator
        #
        # Return the cursor: let it be its own iterable. This could be a
        # very large result set. For things that matter, like gevent,
        # consume in batches allowing periodic switches.
        if prev_polled_tid is None:
            # This is the first time the connection has polled.
            # We'd have to list the entire database for the changes,
            # which is clearly no good, so we want to fetch just the newest
            # TID.
            return None, self.get_current_tid(cursor)

        params = {'tid': prev_polled_tid}
        self._poll_inv_query.execute(cursor, params)
        rows = cursor.fetchall()
        if not rows:
            if self.transactions_may_go_backwards:
                # No detectable changes. Perhaps we went backwards? Check that,
                # but only if it's a possibility
                self._poll_newest_tid_query.execute(cursor)
                new_polled_tid = self.get_current_tid(cursor)
            else:
                # Assume we're fully caught up and that transactions cannot
                # go back
                new_polled_tid = prev_polled_tid
        else:
            new_polled_tid = rows[0][1]

        if new_polled_tid == prev_polled_tid:
            # No transactions have been committed since prev_polled_tid.
            return (), new_polled_tid

        if new_polled_tid < prev_polled_tid:
            # The database connection is stale. This can happen after
            # reading an asynchronous slave that is not fully up to date.
            # (It may also suggest that transaction IDs are not being created
            # in order, which would be a serious bug leading to consistency
            # violations.)
            if self.revert_when_stale:
                # This client prefers to revert to the old state.
                log.warning(
                    "Reverting to stale transaction ID %d and clearing cache. "
                    "(prev_polled_tid=%d)",
                    new_polled_tid, prev_polled_tid)
                # We have to invalidate the whole cPickleCache, otherwise
                # the cache would be inconsistent with the reverted state.
                return None, new_polled_tid

            # This client never wants to revert to stale data, so
            # raise ReadConflictError to trigger a retry.
            # We're probably just waiting for async replication
            # to catch up, so retrying could do the trick.
            raise StaleConnectionError.from_prev_and_new_tid(
                prev_polled_tid, new_polled_tid)

        # New transaction(s) have been added.

        # In the past, but only for history-preserving databases, we
        # would check to see if the previously polled transaction no
        # longer exists in the transaction table. If it didn't, we
        # would return ``(None, new_polled_tid)``, in order to clear
        # the Connection cache.
        #
        # However, we ran for yers without an analogous case for
        # history-free databases without problems, on the theory that
        # all the unreachable objects will be garbage collected
        # anyway.
        #
        # Thus we became convinced it was safe to remove the check in
        # history-preserving databases.
        return rows, new_polled_tid



#!/usr/bin/env python
##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
ZODB storage conversion utility.
"""
from __future__ import print_function

import argparse
import logging
import sys
from io import StringIO

import ZConfig
from persistent.timestamp import TimeStamp
from ZODB.utils import p64
from ZODB.utils import readable_tid_repr
from ZODB.utils import u64

schema_xml = u"""
<schema>
  <import package="ZODB"/>
  <import package="relstorage"/>
  <section type="ZODB.storage" name="source" attribute="source"
    required="yes" />
  <section type="ZODB.storage" name="destination" attribute="destination"
    required="yes" />
</schema>
"""

log = logging.getLogger("zodbconvert")


def storage_has_data(storage):
    i = storage.iterator()
    try:
        try:
            next(i)
        except (IndexError, StopIteration):
            return False
        return True
    finally:
        if hasattr(i, 'close'):
            i.close()

class _DefaultStartStorageIteration(object):
    # At IStorageIteration instance that keeps a default start value.
    # This is needed because RelStorage.iterator() does return an object with an
    # iterator() method, but that object returns itself, so it can only be iterated
    # once! This breaks some implementations of copyTransactionsFrom, notably
    # our own. See #22

    def __init__(self, source, start):
        self._source = source
        self._start = start

    def iterator(self, start=None, end=None):
        return self._source.iterator(start or self._start, end)

    def __getattr__(self, name):
        return getattr(self._source, name)

def open_storages(options):
    schema = ZConfig.loadSchemaFile(StringIO(schema_xml))
    config, _ = ZConfig.loadConfigFile(schema, options.config_file)
    source = config.source.open()
    destination = config.destination.open()

    return source, destination

def main(argv=None):
    # pylint:disable=too-many-branches,too-many-statements
    if argv is None:
        argv = sys.argv
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--dry-run", dest="dry_run", action="store_true",
        default=False,
        help="Attempt to open both storages, then explain what would be done.")
    parser.add_argument(
        "--clear", dest="clear", action="store_true",
        default=False,
        help="Clear the contents of the destination storage before copying."
             " Only works if the destination is a RelStorage."
             " WARNING: use this only if you are certain the destination has no useful data.")
    parser.add_argument(
        "--incremental", dest="incremental", action="store_true",
        help="Assume the destination contains a partial copy of the source "
             "and resume copying from the last transaction. WARNING: no "
             "effort is made to verify that the destination holds the same "
             "transaction data before this point! Use at your own risk. ")
    parser.add_argument(
        '--debug', dest="debug", action='store_true',
        default=False,
        help="Set the logging level to DEBUG instead of the default of "
             "INFO."
    )
    parser.add_argument("config_file", type=argparse.FileType('r'))

    options = parser.parse_args(argv[1:])

    logging.basicConfig(
        level=logging.DEBUG if options.debug else logging.INFO,
        format="%(asctime)s [%(name)s] %(levelname)s %(message)s"
    )

    source, destination = open_storages(options)

    def cleanup_and_exit(exit_msg=None):
        source.close()
        destination.close()
        if exit_msg:
            sys.exit(msg)

    log.info("Storages opened successfully.")

    if options.incremental:
        assert hasattr(destination, 'lastTransaction'), (
            "Error: no API is known for determining the last committed "
            "transaction of the destination storage. Aborting "
            "conversion.")

        if not storage_has_data(destination):
            log.warning("Destination empty, start conversion from the beginning.")
        else:
            # This requires that the storage produce a valid (not z64) value before
            # anything is loaded with it.
            last_tid = destination.lastTransaction()
            if isinstance(last_tid, bytes):
                # This *should* be a byte string.
                last_tid = u64(last_tid)

            next_tid = p64(last_tid + 1)
            # Compensate for the RelStorage bug(?) and get a reusable iterator
            # that starts where we want it to. There's no harm in wrapping it for
            # other sources like FileStorage too.
            source = _DefaultStartStorageIteration(source, next_tid)
            log.info("Resuming ZODB copy from %s", readable_tid_repr(next_tid))


    if options.dry_run:
        log.info("Dry run mode: not changing the destination.")
        if storage_has_data(destination):
            log.warning("The destination storage has data.")
        count = 0
        for txn in source.iterator():
            log.info('%s user=%s description=%s',
                     TimeStamp(txn.tid), txn.user, txn.description)
            count += 1
        log.info("Would copy %d transactions.", count)
        cleanup_and_exit()
    else:
        if options.clear:
            log.info("Clearing old data...")
            if hasattr(destination, 'zap_all'):
                destination.zap_all()
            else:
                msg = ("Error: no API is known for clearing this type "
                       "of storage. Use another method.")
                cleanup_and_exit(msg)
            log.info("Done clearing old data.")

        if storage_has_data(destination) and not options.incremental:
            msg = "Error: the destination storage has data.  Try --clear."
            cleanup_and_exit(msg)

        destination.copyTransactionsFrom(source)
        cleanup_and_exit()


if __name__ == '__main__':
    main()



##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Mixin to the storage for legacy methods that aren't
needed or don't do anything anymore, but which still need to
exist for interface compliance or simply legacy code bases..

This keeps the main code uncluttered.

"""
from __future__ import absolute_import
from __future__ import print_function

import warnings

class LegacyMethodsMixin(object):

    __slots__ = (
    )

    # BaseStorage (and thus FileStorage) and MappingStorage define
    # tpc_transaction and _transaction, even though that's not part of
    # the interface. We only keep them for compatibility with
    # BaseStorage
    def tpc_transaction(self):
        return self._tpc_phase.transaction

    # For tests
    _transaction = property(tpc_transaction)

    def loadEx(self, oid, version=''):
        # Since we don't support versions, just tack the empty version
        # string onto load's result.
        return self.load(oid, version) + ("",)

    def cleanup(self):
        pass

    def supportsVersions(self):
        return False

    def modifiedInVersion(self, oid):
        # pylint:disable=unused-argument
        return ''

    @property
    def _load_cursor(self):
        warnings.warn("_load_cursor is deprecated", FutureWarning, stacklevel=2)
        return self._load_connection.cursor

    @property
    def _load_conn(self):
        warnings.warn("_load_conn is deprecated", FutureWarning, stacklevel=2)
        return self._load_connection.connection

    @property
    def _load_transaction_open(self):
        warnings.warn("_load_transaction_open is deprecated", FutureWarning, stacklevel=2)
        return bool(self._load_connection)

    @property
    def _store_cursor(self):
        warnings.warn("_store_cursor is deprecated", FutureWarning, stacklevel=2)
        return self._store_connection.cursor

    @property
    def _store_conn(self):
        warnings.warn("_store_conn is deprecated", FutureWarning, stacklevel=2)
        return self._store_connection.connection

    def _drop_load_connection(self):
        warnings.warn("_drop_load_connection is deprecated", FutureWarning, stacklevel=2)
        self._load_connection.drop()


    @property
    def fshelper(self):
        warnings.warn("fshelper is deprecated", FutureWarning, stacklevel=2)
        return self.blobhelper.fshelper



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""A foundation for history-preserving RelStorage tests"""
import unittest

import transaction
from persistent.mapping import PersistentMapping

from ZODB.Connection import TransactionMetaData
from ZODB.DB import DB
from ZODB.serialize import referencesf
from ZODB.POSException import POSKeyError
from ZODB.tests import HistoryStorage
from ZODB.tests import IteratorStorage
from ZODB.tests import PackableStorage
from ZODB.tests import RevisionStorage
from ZODB.tests import TransactionalUndoStorage
from ZODB.tests.MinPO import MinPO
from ZODB.tests.StorageTestBase import zodb_pickle
# This class is sadly not cooperative with its superclass,
# so we need to explicitly place it at the back of the MRO.
from ZODB.tests.util import TestCase as ZODBTestCase

from ZODB.utils import p64

from relstorage.storage import bytes8_to_int64
from relstorage.tests.RecoveryStorage import UndoableRecoveryStorage
from relstorage.tests.reltestbase import GenericRelStorageTests
from relstorage.tests.reltestbase import AbstractFromFileStorage
from relstorage.tests.reltestbase import AbstractToFileStorage

# Prior to https://github.com/zopefoundation/ZODB/pull/281, ZODB's
# tests had some previous transaction ids as native strings, not
# bytes. Correct that.
RevisionStorage.ZERO = b'\0' * 8
TransactionalUndoStorage.ZERO = b'\0' * 8

class HistoryPreservingRelStorageTests(GenericRelStorageTests,
                                       TransactionalUndoStorage.TransactionalUndoStorage,
                                       IteratorStorage.IteratorStorage,
                                       IteratorStorage.ExtendedIteratorStorage,
                                       RevisionStorage.RevisionStorage,
                                       PackableStorage.PackableUndoStorage,
                                       HistoryStorage.HistoryStorage,
                                       ZODBTestCase):
    # pylint:disable=too-many-ancestors,abstract-method,too-many-locals,too-many-public-methods
    keep_history = True

    def checkUndoMultipleConflictResolution(self, *_args, **_kwargs): # pylint:disable=signature-differs
        # pylint:disable=arguments-differ
        # 4.2.3 and above add this. it's an exotic feature according to jimfulton.
        raise unittest.SkipTest("conflict-resolving undo not supported")

    def checkIterationIntraTransaction(self):
        # XXX: This test overrides the broken version from
        # IteratorStorage; prior to
        # https://github.com/zopefoundation/ZODB/pull/281 it passed a
        # native str, not bytes, as the previous tid.
        oid = self._storage.new_oid()
        t = TransactionMetaData()
        data = zodb_pickle(MinPO(0))
        try:
            self._storage.tpc_begin(t)
            self._storage.store(oid, RevisionStorage.ZERO, data, '', t)
            self._storage.tpc_vote(t)
            # Don't do tpc_finish yet
            it = self._storage.iterator()
            for x in it:
                self.assertIsNotNone(x)
        finally:
            self._storage.tpc_finish(t)

    def checkTransactionalUndoIterator(self):
        # this test overrides the broken version in TransactionalUndoStorage.

        s = self._storage

        BATCHES = 4
        OBJECTS = 4

        orig = []
        for i in range(BATCHES):
            t = transaction.Transaction()
            tid = p64(i + 1)
            s.tpc_begin(t, tid)
            for j in range(OBJECTS):
                oid = s.new_oid()
                obj = MinPO(i * OBJECTS + j)
                s.store(oid, None, zodb_pickle(obj), '', t)
                orig.append((tid, oid))
            s.tpc_vote(t)
            s.tpc_finish(t)

        # pylint:disable=unnecessary-comprehension
        orig = [(tid, oid, s.getTid(oid)) for tid, oid in orig]

        i = 0
        for tid, oid, revid in orig:
            self._dostore(oid, revid=revid, data=MinPO(revid),
                          description="update %s" % i)

        # Undo the OBJECTS transactions that modified objects created
        # in the ith original transaction.

        def undo(i):
            info = s.undoInfo()
            t = transaction.Transaction()
            s.tpc_begin(t)
            base = i * OBJECTS + i
            for j in range(OBJECTS):
                tid = info[base + j]['id']
                s.undo(tid, t)
            s.tpc_vote(t)
            s.tpc_finish(t)

        for i in range(BATCHES):
            undo(i)

        # There are now (2 + OBJECTS) * BATCHES transactions:
        #     BATCHES original transactions, followed by
        #     OBJECTS * BATCHES modifications, followed by
        #     BATCHES undos

        iter = s.iterator()
        offset = 0

        eq = self.assertEqual

        for i in range(BATCHES):
            txn = iter[offset]
            offset += 1

            tid = p64(i + 1)
            eq(txn.tid, tid)

            L1 = [(rec.oid, rec.tid, rec.data_txn) for rec in txn]
            L2 = [(oid, revid, None) for _tid, oid, revid in orig
                  if _tid == tid]

            eq(L1, L2)

        for i in range(BATCHES * OBJECTS):
            txn = iter[offset]
            offset += 1
            eq(len([rec for rec in txn if rec.data_txn is None]), 1)

        for i in range(BATCHES):
            txn = iter[offset]
            offset += 1

            # The undos are performed in reverse order.
            otid = p64(BATCHES - i)
            L1 = [rec.oid for rec in txn]
            L2 = [oid for _tid, oid, revid in orig if _tid == otid]
            L1.sort()
            L2.sort()
            eq(L1, L2)

        self.assertRaises(IndexError, iter.__getitem__, offset)

    def checkNonASCIITransactionMetadata(self):
        # Verify the database stores and retrieves non-ASCII text
        # in transaction metadata.
        ugly_string = ''.join(chr(c) for c in range(256))
        if isinstance(ugly_string, bytes):
            # Always text. Use latin 1 because it can decode any arbitrary
            # bytes.
            ugly_string = ugly_string.decode('latin-1')

        # The storage layer is defined to take bytes (implicitly in
        # older ZODB releases, explicitly in ZODB 5.something), but historically
        # it can accept either text or bytes. However, it always returns bytes
        check_string = ugly_string.encode("utf-8")

        db = DB(self._storage)
        try:
            c1 = db.open()
            r1 = c1.root()
            r1['alpha'] = 1
            transaction.get().setUser(ugly_string)
            transaction.commit()
            r1['alpha'] = 2
            transaction.get().note(ugly_string)
            transaction.commit()

            info = self._storage.undoInfo()
            __traceback_info__ = info
            self.assertEqual(info[0]['description'], check_string)
            self.assertEqual(info[1]['user_name'], b'/ ' + check_string)
        finally:
            db.close()

    def checkPackGC(self, expect_object_deleted=True, close=True):
        db = DB(self._storage)
        try:
            c1 = db.open()
            r1 = c1.root()
            r1['alpha'] = PersistentMapping()
            transaction.commit()

            oid = r1['alpha']._p_oid
            r1['alpha'] = None
            transaction.commit()

            # The object should still exist
            self._storage.load(oid, '')

            # Pack
            self._storage.pack(self._storage.lastTransactionInt(), referencesf)
            self._storage.sync()

            if expect_object_deleted:
                # The object should now be gone
                self.assertRaises(KeyError, self._storage.load, oid, '')
            else:
                # The object should still exist
                self._storage.load(oid, '')
        finally:
            if close:
                db.close()
        return oid

    def checkPackGCDisabled(self):
        self._storage = self.make_storage(pack_gc=False)
        self.checkPackGC(expect_object_deleted=False)

    def checkPackGCPrePackOnly(self):
        self._storage = self.make_storage(pack_prepack_only=True)
        self.checkPackGC(expect_object_deleted=False)

    def checkPackGCReusePrePackData(self):
        self._storage = self.make_storage(pack_prepack_only=True)
        oid = self.checkPackGC(expect_object_deleted=False, close=False)
        # We now have pre-pack analysis data
        self._storage._options.pack_prepack_only = False
        self._storage.pack(0, referencesf, skip_prepack=True)
        # The object should now be gone
        self.assertRaises(KeyError, self._storage.load, oid, '')
        self._storage.close()

    def checkPackOldUnreferenced(self):
        db = DB(self._storage)
        try:
            c1 = db.open()
            r1 = c1.root()
            r1['A'] = PersistentMapping()
            A_B = PersistentMapping()
            r1['A']['B'] = A_B
            transaction.get().note(u'add A then add B to A')
            transaction.commit()

            del r1['A']['B']
            transaction.get().note(u'remove B from A')
            transaction.commit()

            r1['A']['C'] = ''
            transaction.get().note(u'add C (non-persistent) to A')
            transaction.commit()

            packtime = c1._storage.lastTransactionInt()
            self._storage.pack(packtime, referencesf)

            # B should be gone, since nothing refers to it.
            with self.assertRaises(KeyError):
                __traceback_info__ = bytes8_to_int64(A_B._p_oid)
                self._storage.load(A_B._p_oid)

        finally:
            db.close()

    def checkHistoricalConnection(self):
        import persistent
        import ZODB.POSException
        db = DB(self._storage)
        conn = db.open()
        root = conn.root()

        root['first'] = persistent.mapping.PersistentMapping(count=0)
        transaction.commit()

        time_of_first_transaction = conn._storage.lastTransaction()

        root['second'] = persistent.mapping.PersistentMapping()
        root['first']['count'] += 1
        transaction.commit()

        transaction1 = transaction.TransactionManager(explicit=True)

        historical_conn = db.open(transaction_manager=transaction1, at=time_of_first_transaction)

        eq = self.assertEqual

        # regular connection sees present:

        eq(sorted(conn.root().keys()), ['first', 'second'])
        eq(conn.root()['first']['count'], 1)

        # historical connection sees past:
        transaction1.begin()
        eq(sorted(historical_conn.root().keys()), ['first'])
        eq(historical_conn.root()['first']['count'], 0)

        # Can't change history:

        historical_conn.root()['first']['count'] += 1
        eq(historical_conn.root()['first']['count'], 1)
        self.assertRaises(ZODB.POSException.ReadOnlyHistoryError,
                          transaction1.commit)
        transaction1.abort()
        eq(historical_conn.root()['first']['count'], 0)

        # Making a change in the present
        root['third'] = 3
        transaction.commit()

        # Is also not reflected in the past, even after explicit sync,
        transaction1.begin()
        eq(sorted(historical_conn.root().keys()), ['first'])
        eq(historical_conn.root()['first']['count'], 0)
        # Since we cannot change anything, we cannot join a transaction either.
        # The afterCompletion call is never invoked.
        historical_conn._storage._storage.afterCompletion = lambda: self.fail("Not called")
        transaction1.commit()


        historical_conn.close()
        conn.close()
        db.close()

    def checkMigrateTransactionEmpty(self):
        # The transaction.empty column gets renamed in 'prepare'
        adapter = self._storage._adapter
        schema = adapter.schema
        # First, we have to flip it back to the old name, since we installed
        # with the name we want.
        stmt = schema._rename_transaction_empty_stmt
        # ALTER TABLE transaction RENAME empty TO is_empty
        # or
        # ALTER TABLE transaction CHANGE empty is_empty
        stmt = stmt.replace('is_empty', 'FOOBAR')
        stmt = stmt.replace('empty', 'is_empty')
        stmt = stmt.replace("FOOBAR", 'empty')

        test_conn, test_cursor = adapter.connmanager.open()
        try:

            try:
                test_cursor.execute(stmt)
            except Exception as e:
                # XXX: This should be more strict. We really just
                # want to catch the db-api specific ProgrammingError,
                # and only on MySQL 8.0+. But we don't have a good way to do that.
                raise unittest.SkipTest(str(e))

            self.assertTrue(schema._needs_transaction_empty_update(test_cursor))

            schema.update_schema(test_cursor, None)
            self.assertFalse(schema._needs_transaction_empty_update(test_cursor))
        finally:
            adapter.connmanager.close(test_conn, test_cursor)

    def __setup_checkImplementsIExternalGC(self):
        from zope.interface.verify import verifyObject
        import ZODB.interfaces
        verifyObject(ZODB.interfaces.IExternalGC, self._storage)

        # Now do it.
        # We need to create a few different revisions of an object
        # so that we can selectively remove old versions and check that
        # when we remove the final version, the whole thing goes away.
        db = self._closing(ZODB.DB(self._storage))
        conn = self._closing(db.open())
        root = conn.root()
        root['key'] = PersistentMapping()
        transaction.commit()

        for i in range(5):
            tx = transaction.begin()
            tx.description = u'Revision %s' % i
            root['key']['item'] = i
            transaction.commit()

        obj_oid = root['key']._p_oid

        return db, conn, obj_oid

    def checkImplementsIExternalGC(self):
        db, conn, obj_oid = self.__setup_checkImplementsIExternalGC()

        storage = conn._storage
        history = storage.history(obj_oid, size=100)
        self.assertEqual(6, len(history))
        latest_tid = history[0]['tid']
        # We can delete the latest TID for the OID, and the whole
        # object goes away on a pack.
        t = TransactionMetaData()
        storage.tpc_begin(t)
        count = storage.deleteObject(obj_oid, latest_tid, t)
        self.assertEqual(count, 1)
        # Doing it again will do nothing because it's already
        # gone.
        count = storage.deleteObject(obj_oid, latest_tid, t)
        invalidations = storage.tpc_vote(t)
        storage.tpc_finish(t)

        # It is detected as invalidated during voting
        self.assertEqual([obj_oid], list(invalidations))

        # And not at the next poll.
        invalidations = storage.poll_invalidations()
        self.assertEqual([], list(invalidations))

        # Getting the most recent fails.
        with self.assertRaises(POSKeyError):
            storage.load(obj_oid)

        # But we can load a state before then.
        state = storage.loadSerial(obj_oid, history[1]['tid'])
        self.assertEqual(len(state), history[1]['size'])

        # Length hasn't changed
        storage._adapter.stats.large_database_change()
        self.assertEqual(len(storage), 2)

        # History records it as a deletion with size 0
        history_after = storage.history(obj_oid)
        self.assertEqual(0, history_after[0]['size'])

        # Now if we proceed to pack it, *without* doing a GC...
        from relstorage.storage.pack import Pack
        options = storage._options.copy(pack_gc=False)
        self.assertFalse(options.pack_gc)
        packer = Pack(options, storage._adapter, storage.blobhelper, storage._cache)
        self.assertFalse(packer.options.pack_gc)
        packer.pack(storage.lastTransactionInt(), referencesf)

        # ... and bring the storage into the current view...
        storage.sync()

        # ...then the object is gone in all revisions...
        with self.assertRaises(POSKeyError):
            storage.load(obj_oid)

        for history_item in history:
            tid = history_item['tid']
            with self.assertRaises(POSKeyError):
                storage.loadSerial(obj_oid, tid)

        # ...and the size is smaller.
        self.assertEqual(len(storage), 1)
        conn.close()
        db.close()

    ###
    # Fixes for tests that assume the current clock
    # and the TID clock are the same.
    ###

    def __tid_clock_needs_care(self):
        adapter = self._storage._adapter
        return getattr(adapter, 'RS_TEST_TXN_PACK_NEEDS_SLEEP', False)

    def __maybe_ignore_monotonic(self, cls, method_name):
        if not self.__tid_clock_needs_care():
            return getattr(super(HistoryPreservingRelStorageTests, self), method_name)()

        # Override one from RevisionStorage to go back to actually sleeping,
        # since our TID clock is external now.
        unbound = getattr(cls, method_name)
        if hasattr(unbound, 'im_func'):
            # We're on python 2. There's no __wrapped__ to give us access to the raw
            # function, we have to dig it out by hand. Our only option is the closure.
            function_wrapper = unbound.im_func
            unbound = function_wrapper.__closure__[0].cell_contents
        else:
            unbound = unbound.__wrapped__ # pylint:disable=no-member

        bound = lambda: unbound(self)
        self.__never_snoozing(bound)

    def __never_snoozing(self, method):
        def never_snooze():
            return

        before_snooze = RevisionStorage.snooze
        assert PackableStorage.snooze == RevisionStorage.snooze
        RevisionStorage.snooze = never_snooze
        PackableStorage.snooze = never_snooze
        try:
            return method()
        finally:
            RevisionStorage.snooze = before_snooze
            PackableStorage.snooze = before_snooze

    def checkLoadBefore(self):
        # Most of the time this works, but sometimes it fails an internal assertion,
        # most commonly seen on AppVeyor.
        # https://ci.appveyor.com/project/jamadden/relstorage/builds/26243441/job/p24ocr2ir6wpvg3v#L1087
        raise unittest.SkipTest("Assumes it can control timestamps")

    def checkLoadBeforeOld(self):
        self.__maybe_ignore_monotonic(RevisionStorage.RevisionStorage,
                                      'checkLoadBeforeOld')

    def checkSimpleHistory(self):
        if not self.__tid_clock_needs_care():
            return super(HistoryPreservingRelStorageTests, self).checkSimpleHistory()
        # This assumes that the `time` value in the storage.history()
        # for an object always increases, even though there are 8-byte TID values
        # that, while themselves increasing, round down to equal floating point
        # time values. For example, these two values are in the proper sequence:
        #   b'\x03\xd1K\xc6-\xf33!' == 275084366792831777,
        # and
        #   b'\x03\xd1K\xc6-\xf33"' == 275084366792831778,
        # But both have
        #   TimeStamp(tid).timeTime() == 1563997810.769531.
        #
        # This test tries to do something about that (delaying between transactions to let
        # time.time() move forward), but only on Windows, and it turns out even if we
        # let it sleep it doesn't actually help things 100% of the time (99% yes, but not
        # 100%), so we took that out. See below.

        # Even wheen sleeping, we've seen a very small number of apparent genuine failures
        # (Python 3.5 on Travis), that of reporting
        #
        #      HistoryStorage.py", line 54, in _checkHistory
        #        self.assertLess(a, b)
        #    AssertionError:  1564151039.9015017 not less than 1564150979.908543
        #
        # Which is true, it isn't. In fact, there's a difference of
        # 59.99s between the two stamps (all the errors have shown that same amount).
        #
        # - 1564151039.9015017 -> '2019-07-26 14:23:59.901502' -> 275095335361341917
        # - 1564150979.9085430 -> '2019-07-26 14:22:59.908543' -> 275095331066878668
        #
        # And
        # - a -> \x03\xd1U\xbf\xff\x94i\xdd    -> [3, 209, 85, 191, 255, 148, 105, 221]
        # - b -> \x03\xd1U\xbe\xff\x9c\x1a\xcc -> [3, 209, 85, 190, 255, 156,  26, 204]
        #
        # Which makes very little sense:
        #
        # - The DB query orders by TID, DESC;
        # - We iterate those in that same order.
        #
        # But the very first comparison is against the local value of
        # time.time(); time.time() isn't guaranteed to be monotonic
        # increasing and the docs specifically say it can move
        # backwards, so perhaps it did?
        #
        # I was able to catch an instance of this locally running
        # ``ztest --layer My -t checkSimpleHistory --repeat 15`` and
        # indeed, the failure was against the local time.time() value
        # on the first comparison.
        #
        # The local time was              1564152719.998203
        # The DB TID (in the future) was  275095459921661457 -> '\x03\xd1U\xdc\xff\xf4\x0e\x11' ->
        #                                 1564152779.9890637
        # Note the DB TID is one of those ambiguous ones, plugging the unix timestamp
        # back into the process we get:
        #
        #   '\x03\xd1U\xdc\xff\xf4\x0e\x00' -> 275095459921661440
        #
        # The previous TID in the database was
        #   '\x03\xd1U\xdc\xff\xecD\xaa'    -> 275095459921151146
        #
        # Which is a much greater difference than would be needed (i.e., we didn't have to add 1 to
        # get to the new tid).
        #
        # So the only conclusion can be that we just can't compare
        # local time directly to database time.
        #
        # We fix this by disabling 'assertLess'; the test contains
        # calls to assertEqual() on the TIDs, which makes much more
        # sense anyway, so we don't lose much.

        self.assertLess = lambda *args: None

        try:
            super(HistoryPreservingRelStorageTests, self).checkSimpleHistory()
        finally:
            del self.assertLess

    def _run_with_storage_packing_at_packtime(self,
                                              methname,
                                              find_packtime):
        # Ignore the pack timestamp given. Execute `find_packtime(storage)`
        # instead and use that.
        meth = getattr(super(HistoryPreservingRelStorageTests, self), methname)
        if not self.__tid_clock_needs_care():
            return meth()

        orig_pack = self._storage.pack
        def pack(_ignored_packtime, ref_getter):
            packtime = find_packtime(self._storage)
            orig_pack(packtime, ref_getter)
        self._storage.pack = pack

        try:
            return self.__never_snoozing(meth)
        finally:
            del self._storage.pack
            orig_pack = None

    def _run_packing_to_latest(self, methname):
        # Ignore the pack timestamp given. The test is trying to pack to "now", but
        # because of how fast we commit, that doesn't always work out correctly.
        # Instead, use the actual most recent tid.
        return self._run_with_storage_packing_at_packtime(
            methname,
            find_packtime=lambda s: s.lastTransactionInt())

    def checkPackJustOldRevisions(self):
        self._run_packing_to_latest('checkPackJustOldRevisions')

    def checkPackAllRevisions(self):
        self._run_packing_to_latest('checkPackAllRevisions')

    def checkPackOnlyOneObject(self):
        self._run_packing_to_latest('checkPackOnlyOneObject')

    def checkPackUndoLog(self):
        packafter = []
        orig_dostore = self._dostoreNP
        def dostorenp(*args, **kwargs):
            result = orig_dostore(*args, **kwargs)
            if not packafter:
                packafter.append(self._storage.lastTransactionInt())
            return result
        self._dostoreNP = dostorenp

        def find_packtime(_storage):
            assert len(packafter) == 1, packafter
            return packafter[0]

        try:
            self._run_with_storage_packing_at_packtime('checkPackUndoLog',
                                                       find_packtime)
        finally:
            del self._dostoreNP



    def __pack_after_nth_commit_on_nth_child_instance(self, meth_name,
                                                      commit_number, child_number=0):
        # Use child_number = 0 to specify self._storage.
        # Currently only support that or the first child
        assert child_number in (0, 1)

        commit_tids = []
        def watching_tpc_finish(inst, orig_finish, *args, **kwargs):
            tid = orig_finish(*args, **kwargs)
            commit_tids.append(tid)
            if len(commit_tids) == commit_number:
                del inst.tpc_finish
            return tid

        s = self._storage
        if child_number == 1:
            orig_new_instance = s.new_instance
            def new_instance():
                inst = orig_new_instance()
                orig_finish = inst.tpc_finish
                inst.tpc_finish = lambda *args, **kw: watching_tpc_finish(inst, orig_finish,
                                                                          *args, **kw)
                return inst
            s.new_instance = new_instance
        else:
            orig_finish = s.tpc_finish
            s.tpc_finish = lambda *args, **kw: watching_tpc_finish(s, orig_finish, *args, **kw)

        def find_packtime(_storage):
            assert len(commit_tids) == commit_number
            return bytes8_to_int64(commit_tids[commit_number - 1])

        try:
            self._run_with_storage_packing_at_packtime(meth_name, find_packtime)
        finally:
            # Break cycles
            s.new_instance = None
            s = None

    def __pack_after_first_commit_on_first_child_instance(self, meth_name):
        return self.__pack_after_nth_commit_on_nth_child_instance(
            meth_name,
            1,
            1
        )

    def checkTransactionalUndoAfterPackWithObjectUnlinkFromRoot(self):
        self.__pack_after_first_commit_on_first_child_instance(
            'checkTransactionalUndoAfterPackWithObjectUnlinkFromRoot'
        )

    def checkTransactionalUndoAfterPack(self):
        self.__pack_after_nth_commit_on_nth_child_instance(
            'checkTransactionalUndoAfterPack',
            2,
            0
        )

    def checkPackUnlinkedFromRoot(self):
        self.__pack_after_first_commit_on_first_child_instance('checkPackUnlinkedFromRoot')

class HistoryPreservingToFileStorage(AbstractToFileStorage,
                                     UndoableRecoveryStorage,
                                     ZODBTestCase):
    # pylint:disable=too-many-ancestors,abstract-method,too-many-locals
    keep_history = True


class HistoryPreservingFromFileStorage(AbstractFromFileStorage,
                                       UndoableRecoveryStorage,
                                       ZODBTestCase):
    # pylint:disable=too-many-ancestors,abstract-method,too-many-locals
    keep_history = True
